from __future__ import annotations


__copyright__ = "Copyright (C) 2016 Matt Wala"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from loopy.diagnostic import LoopyError
from loopy.schedule import (
    Barrier,
    CallKernel,
    EnterLoop,
    ReturnFromKernel,
    RunInstruction,
)
from loopy.schedule.tools import get_block_boundaries


def map_schedule_onto_host_or_device(kernel):
    # FIXME: Should be idempotent.
    from loopy.kernel import KernelState
    assert kernel.state == KernelState.LINEARIZED

    from functools import partial
    device_prog_name_gen = partial(
                kernel.get_var_name_generator(),
                kernel.target.device_program_name_prefix
                + kernel.name
                + kernel.target.device_program_name_suffix)

    if not kernel.target.split_kernel_at_global_barriers():
        new_schedule = (
            [CallKernel(kernel_name=device_prog_name_gen()),
                *kernel.linearization,
                ReturnFromKernel(kernel_name=kernel.name)])
        kernel = kernel.copy(linearization=new_schedule)
    else:
        kernel = map_schedule_onto_host_or_device_impl(
                kernel, device_prog_name_gen)

    return kernel


def map_schedule_onto_host_or_device_impl(kernel, device_prog_name_gen):
    schedule = kernel.linearization
    loop_bounds = get_block_boundaries(schedule)

    # {{{ inner mapper function

    dummy_call = CallKernel(kernel_name="")
    dummy_return = ReturnFromKernel(kernel_name="")

    def inner_mapper(start_idx, end_idx, new_schedule):
        schedule_required_splitting = False

        i = start_idx
        current_chunk = []
        while i <= end_idx:
            sched_item = schedule[i]

            if isinstance(sched_item, RunInstruction):
                current_chunk.append(sched_item)
                i += 1

            elif isinstance(sched_item, EnterLoop):
                loop_end = loop_bounds[i]
                inner_schedule = []
                loop_required_splitting = inner_mapper(
                    i + 1, loop_end - 1, inner_schedule)

                start_item = schedule[i]
                end_item = schedule[loop_end]

                i = loop_end + 1

                if loop_required_splitting:
                    schedule_required_splitting = True
                    if current_chunk:
                        new_schedule.extend(
                            [dummy_call.copy(), *current_chunk, dummy_return.copy()])
                    new_schedule.extend(
                        [start_item, *inner_schedule, end_item])
                    current_chunk = []
                else:
                    current_chunk.extend(
                        [start_item, *inner_schedule, end_item])

            elif isinstance(sched_item, Barrier):
                if sched_item.synchronization_kind == "global":
                    # Wrap the current chunk into a kernel call.
                    schedule_required_splitting = True
                    if current_chunk:
                        new_schedule.extend(
                            [dummy_call.copy(), *current_chunk, dummy_return.copy()])
                    new_schedule.append(sched_item)
                    current_chunk = []
                else:
                    current_chunk.append(sched_item)
                i += 1
            else:
                raise LoopyError("unexpected type of schedule item: %s"
                        % type(sched_item).__name__)

        if current_chunk and schedule_required_splitting:
            # Wrap remainder of schedule into a kernel call.
            new_schedule.extend(
                [dummy_call.copy(), *current_chunk, dummy_return.copy()])
        else:
            new_schedule.extend(current_chunk)

        return schedule_required_splitting

    # }}}

    new_schedule = []
    split_kernel = inner_mapper(0, len(schedule) - 1, new_schedule)
    if not split_kernel:
        # Wrap everything into a kernel call.
        new_schedule = (
            [dummy_call.copy(), *new_schedule, dummy_return.copy()])

    # Assign names to CallKernel / ReturnFromKernel instructions

    for idx, sched_item in enumerate(new_schedule):
        if isinstance(sched_item, CallKernel):
            last_kernel_name = device_prog_name_gen()
            new_schedule[idx] = sched_item.copy(kernel_name=last_kernel_name)
        elif isinstance(sched_item, ReturnFromKernel):
            new_schedule[idx] = sched_item.copy(kernel_name=last_kernel_name)

    new_kernel = kernel.copy(linearization=new_schedule)

    return new_kernel



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from typing import Any, Mapping, Type, Union

import numpy as np
from typing_extensions import TypeAlias

from loopy.diagnostic import LoopyError
from loopy.typing import auto


__doc__ = """
.. currentmodule:: loopy

.. autoclass:: LoopyType

.. autoclass:: ToLoopyTypeConvertible

.. autoclass:: NumpyType

.. currentmodule:: loopy.types

.. autoclass:: AtomicType

.. autoclass:: AtomicNumpyType
"""


class LoopyType:
    """
    Abstract class for dtypes of variables encountered in a
    :class:`loopy.LoopKernel`.
    """
    def is_integral(self) -> bool:
        raise NotImplementedError()

    def is_complex(self) -> bool:
        raise NotImplementedError()

    def uses_complex(self) -> bool:
        raise NotImplementedError()

    def is_composite(self) -> bool:
        raise NotImplementedError()

    @property
    def itemsize(self) -> int:
        raise NotImplementedError()

    @property
    def numpy_dtype(self) -> np.dtype:
        raise ValueError("'%s' is not a numpy type"
                % str(self))


class AtomicType(LoopyType):
    """
    Abstract class for dtypes of variables encountered in a :class:`loopy.LoopKernel`
    on which atomic operations are performed .
    """


# {{{ numpy-based dtype

class NumpyType(LoopyType):
    def __init__(self, dtype: np.dtype):
        assert not isinstance(dtype, LoopyType)

        if dtype is None:
            raise TypeError("may not pass None to construct NumpyType")

        if dtype == object:  # noqa: E721
            raise TypeError("loopy does not directly support object arrays")

        self.dtype = np.dtype(dtype)

    def __hash__(self) -> int:
        return hash(self.dtype)

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, self.dtype)

    def __eq__(self, other: object) -> bool:
        return (
                type(self) is type(other)
                # mypy doesn't understand 'type(self) is type(other)'
                and self.dtype == other.dtype)  # type: ignore[attr-defined]

    def is_integral(self) -> bool:
        return self.dtype.kind in "iu"

    def is_complex(self) -> bool:
        return self.dtype.kind == "c"

    def involves_complex(self) -> bool:
        def dtype_involves_complex(dtype):
            if dtype.kind == "c":
                return True

            if dtype.fields is None:
                return False
            else:
                return any(
                        dtype_involves_complex(f[0])
                        for f in dtype.fields.values())

        return dtype_involves_complex(self.dtype)

    def is_composite(self):
        return self.dtype.kind == "V"

    @property
    def itemsize(self) -> int:
        return self.dtype.itemsize

    @property
    def numpy_dtype(self) -> np.dtype:
        return self.dtype

    def __repr__(self) -> str:
        return "np:" + repr(self.dtype)

# }}}


# {{{ atomic dtype

class AtomicNumpyType(NumpyType, AtomicType):
    """A dtype wrapper that indicates that the described type should be capable
    of atomic operations.
    """
    def __hash__(self):
        return 0xa7031c ^ hash(self.dtype)

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, 0xa7031c)
        key_builder.rec(key_hash, self.dtype)

    def __repr__(self):
        return "np_atomic:%s" % repr(self.dtype)

# }}}


# {{{ opaque type

class OpaqueType(LoopyType):
    """An opaque data type is truly opaque - it has no allocations, no
    temporaries of that type, etc. The only thing allowed is to be pass in
    through one ValueArg and go out to another. It is introduced to accommodate
    functional calls to external libraries.
    """
    def __init__(self, name: str) -> None:
        assert isinstance(name, str)
        self.name = name

    def is_integral(self) -> bool:
        return False

    def is_complex(self) -> bool:
        return False

    def involves_complex(self) -> bool:
        return False

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, self.name)

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        return (
                type(self) is type(other)
                # mypy doesn't understand 'type(self) is type(other)'
                and self.name == other.name  # type: ignore[attr-defined]
                )

# }}}


ToLoopyTypeConvertible: TypeAlias = Union[Type[auto], np.dtype, LoopyType, None]


def to_loopy_type(dtype: ToLoopyTypeConvertible,
                  allow_auto: bool = False, allow_none: bool = False,
                  for_atomic: bool = False
                  ) -> type[auto] | LoopyType | None:
    if dtype is None:
        if allow_none:
            return None
        else:
            raise LoopyError("dtype may not be none")

    elif dtype is auto:
        if allow_auto:
            # mypy doesn't seem to catch that this narrows the type of dtype
            return dtype  # type: ignore[return-value]
        else:
            raise LoopyError("dtype may not be auto")

    numpy_dtype = None

    if dtype is not None:
        try:
            # We're playing fast and loose here, and mypy is onto us. It has a
            # point.
            numpy_dtype = np.dtype(dtype)  # type: ignore
        except Exception:
            pass

    if isinstance(dtype, LoopyType):
        if for_atomic:
            if isinstance(dtype, NumpyType):
                return AtomicNumpyType(dtype.dtype)
            elif not isinstance(dtype, AtomicType):
                raise LoopyError("do not know how to convert '%s' to an atomic type"
                        % dtype)

        return dtype

    elif numpy_dtype is not None:
        if for_atomic:
            return AtomicNumpyType(numpy_dtype)
        else:
            return NumpyType(numpy_dtype)

    else:
        raise TypeError("dtype must be a LoopyType, or convertible to one, "
                "found '%s' instead" % type(dtype))


_TO_UNSIGNED_MAPPING: Mapping[np.dtype[Any], np.dtype[Any]] = {
        np.dtype(np.int8): np.dtype(np.uint8),
        np.dtype(np.int16): np.dtype(np.uint16),
        np.dtype(np.int32): np.dtype(np.uint32),
        np.dtype(np.int64): np.dtype(np.uint64),
        }


def to_unsigned_dtype(dtype: np.dtype[Any]) -> np.dtype[Any]:
    if dtype.kind == "u":
        return dtype
    if dtype.kind != "i":
        raise ValueError("can only convert integer types to unsigned")

    return _TO_UNSIGNED_MAPPING[dtype]


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import re
from importlib import metadata


VERSION_TEXT = metadata.version("loopy")
_match = re.match(r"^([0-9.]+)([a-z0-9]*?)$", VERSION_TEXT)
assert _match is not None
VERSION_STATUS = _match.group(2)
VERSION = tuple(int(nr) for nr in _match.group(1).split("."))


try:
    import islpy.version
except ImportError:
    _islpy_version = "_UNKNOWN_"
else:
    _islpy_version = islpy.version.VERSION_TEXT

try:
    import cgen.version
except ImportError:
    _cgen_version = "_UNKNOWN_"
else:
    _cgen_version = cgen.version.VERSION_TEXT

DATA_MODEL_VERSION = f"{VERSION_TEXT}-islpy{_islpy_version}-cgen{_cgen_version}-v2"


FALLBACK_LANGUAGE_VERSION = (2018, 2)
MOST_RECENT_LANGUAGE_VERSION = (2018, 2)

LOOPY_USE_LANGUAGE_VERSION_2018_2 = (2018, 2)

LANGUAGE_VERSION_SYMBOLS = [
        "LOOPY_USE_LANGUAGE_VERSION_2018_2",
        ]

__doc__ = """

.. currentmodule:: loopy
.. data:: VERSION

    A tuple representing the current version number of loopy, for example
    **(2017, 2, 1)**. Direct comparison of these tuples will always yield
    valid version comparisons.

.. _language-versioning:

Loopy Language Versioning
-------------------------

At version 2018.1, :mod:`loopy` introduced a language versioning scheme to make
it easier to evolve the language while retaining backward compatibility. What
prompted this is the addition of
:attr:`loopy.Options.enforce_variable_access_ordered`, which (despite
its name) serves to enable a new check that helps ensure that all variable
access in a kernel is ordered as intended. Since that has the potential to
break existing programs, kernels now have to declare support for a given
language version to let them take advantage of this check.

As a result, :mod:`loopy` will now issue a warning when a call to
:func:`loopy.make_kernel` does not declare a language version. Such kernels
will (indefinitely) default to language version 2018.2.  If passing a
language version to :func:`make_kernel` is impractical, you may also import
one of the ``LOOPY_USE_LANGUAGE_VERSION_...`` symbols given below using::

    from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2

in the global namespace of the function calling :func:`make_kernel`. If
*lang_version* in that call is not explicitly given, this value will be used.

Language versions will generally reflect the version number of :mod:`loopy` in
which they were introduced, though it is likely that most versions of
:mod:`loopy` do not introduce language incompatibilities. In such
situations, the previous language version number remains. (In fact, we
will work hard to avoid backward-incompatible language changes.)

.. data:: MOST_RECENT_LANGUAGE_VERSION

    A tuple representing the most recent language version number of loopy, for
    example **(2018, 1)**. Direct comparison of these tuples will always
    yield valid version comparisons.


History of Language Versions
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. data:: LOOPY_USE_LANGUAGE_VERSION_2018_2

    ``loopy.Options.ignore_boostable_into`` is turned on by default.

.. data:: LOOPY_USE_LANGUAGE_VERSION_2018_1

    :attr:`loopy.Options.enforce_variable_access_ordered` is turned on by
    default. Unsupported from :mod:`loopy` version 2020.2 onwards.

    (no longer available)

.. data:: LOOPY_USE_LANGUAGE_VERSION_2017_2_1

    Initial legacy language version. Unsupported from :mod:`loopy` version
    2020.2 onwards.

    (no longer available)
"""



import numpy as np

import pyopencl as cl
import pyopencl.array

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


ctx = cl.create_some_context()
queue = cl.CommandQueue(ctx)

n = 15 * 10**6
a = cl.array.arange(queue, n, dtype=np.float32)

knl = lp.make_kernel(
        "{ [i]: 0<=i<n }",
        "out[i] = 2*a[i]")

knl = lp.set_options(knl, write_code=True)
knl = lp.split_iname(knl, "i", 4, slabs=(0, 1), inner_tag="vec")
knl = lp.split_array_axis(knl, "a,out", axis_nr=0, count=4)
knl = lp.tag_array_axes(knl, "a,out", "C,vec")

knl(queue, a=a.reshape(-1, 4), n=n)



"""Target for Intel ISPC."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import operator
from functools import reduce
from typing import TYPE_CHECKING, Iterable, Sequence, cast

import numpy as np
from typing_extensions import Never

import pymbolic.primitives as p
from cgen import Collection, Const, Declarator, Generable
from pymbolic import var
from pymbolic.mapper.stringifier import PREC_NONE
from pymbolic.mapper.substitutor import make_subst_func
from pytools import memoize_method

from loopy.diagnostic import LoopyError
from loopy.kernel.data import AddressSpace, ArrayArg, LocalInameTag, TemporaryVariable
from loopy.symbolic import (
    CoefficientCollector,
    CombineMapper,
    GroupHardwareAxisIndex,
    Literal,
    LocalHardwareAxisIndex,
    SubstitutionMapper,
    flatten,
)
from loopy.target.c import CFamilyASTBuilder, CFamilyTarget
from loopy.target.c.codegen.expression import ExpressionToCExpressionMapper


if TYPE_CHECKING:
    from loopy.codegen import CodeGenerationState
    from loopy.codegen.result import CodeGenerationResult
    from loopy.kernel import LoopKernel
    from loopy.kernel.instruction import Assignment
    from loopy.schedule import CallKernel
    from loopy.types import LoopyType
    from loopy.typing import Expression


class IsVaryingMapper(CombineMapper[bool, []]):
    # FIXME: Update this if/when ispc reduction support is added.

    def __init__(self, kernel: LoopKernel) -> None:
        self.kernel = kernel
        super().__init__()

    def combine(self, values: Iterable[bool]) -> bool:
        return reduce(operator.or_, values, False)

    def map_constant(self, expr):
        return False

    def map_group_hw_index(self, expr: GroupHardwareAxisIndex) -> Never:
        # These only exist for a brief blip in time inside the expr-to-cexpr
        # mapper. We should never see them.
        raise AssertionError()

    def map_local_hw_index(self, expr: LocalHardwareAxisIndex) -> Never:
        # These only exist for a brief blip in time inside the expr-to-cexpr
        # mapper. We should never see them.
        raise AssertionError()

    def map_variable(self, expr: p.Variable) -> bool:
        iname = self.kernel.inames.get(expr.name)
        if iname is not None:
            ltags = iname.tags_of_type(LocalInameTag)
            if ltags:
                ltag, = ltags
                assert ltag.axis == 0
                return True

        return False


# {{{ expression mapper

class ExprToISPCExprMapper(ExpressionToCExpressionMapper):
    def _get_index_ctype(self):
        if self.kernel.index_dtype.numpy_dtype == np.int32:
            return "int32"
        elif self.kernel.index_dtype.numpy_dtype == np.int64:
            return "int64"
        else:
            raise ValueError("unexpected index_type")

    def map_group_hw_index(self, expr, type_context):
        return var(
                "((uniform %s) taskIndex%d)"
                % (self._get_index_ctype(), expr.axis))

    def map_local_hw_index(self, expr, type_context):
        if expr.axis == 0:
            return var("(varying %s) programIndex" % self._get_index_ctype())
        else:
            raise LoopyError("ISPC only supports one local axis")

    def map_constant(self, expr, type_context):
        if isinstance(expr, (complex, np.complexfloating)):
            raise NotImplementedError("complex numbers in ispc")
        else:
            if type_context == "f":
                return Literal(repr(float(expr)))
            elif type_context == "d":
                # Keepin' the good ideas flowin' since '66.
                return Literal(repr(float(expr))+"d")
            elif type_context in ["i", "b"]:
                return expr
            else:
                from loopy.typing import is_integer
                if is_integer(expr):
                    return expr

                raise RuntimeError("don't know how to generate code "
                        "for constant '%s'" % expr)

    def map_variable(self, expr, type_context):
        tv = self.kernel.temporary_variables.get(expr.name)

        if tv is not None and tv.address_space == AddressSpace.PRIVATE:
            # FIXME: This is a pretty coarse way of deciding what
            # private temporaries get duplicated. Refine? (See also
            # below in decl generation)
            _gsize, lsize = self.kernel.get_grid_size_upper_bounds_as_exprs(
                    self.codegen_state.callables_table)
            if lsize:
                return expr[var("programIndex")]
            else:
                return expr

        else:
            return super().map_variable(expr, type_context)

    def map_subscript(self, expr, type_context):
        from loopy.kernel.data import TemporaryVariable

        ary = self.find_array(expr)

        if (isinstance(ary, TemporaryVariable)
                and ary.address_space == AddressSpace.PRIVATE):
            # generate access code for access to private-index temporaries

            _gsize, lsize = self.kernel.get_grid_size_upper_bounds_as_exprs()
            if lsize:
                lsize, = lsize
                from pymbolic import evaluate

                from loopy.kernel.array import get_access_info

                access_info = get_access_info(self.kernel, ary, expr.index,
                    lambda expr: evaluate(expr, self.codegen_state.var_subst_map),
                    self.codegen_state.vectorization_info)

                subscript, = access_info.subscripts
                result = var(access_info.array_name)[
                        var("programIndex") + self.rec(lsize*subscript, "i")]

                if access_info.vector_index is not None:
                    return self.kernel.target.add_vector_access(
                        result, access_info.vector_index)
                else:
                    return result

        return super().map_subscript(
                expr, type_context)

    def wrap_in_typecast(self, actual_type: LoopyType, needed_type: LoopyType, s):
        raise NotImplementedError("wrap_in_typecast needs uniform-ness information "
                                  "for ispc")

    def rec(self, expr, type_context=None, needed_type: LoopyType | None = None):  # type: ignore[override]
        result = super().rec(expr, type_context)

        if needed_type is None:
            return result
        else:
            actual_type = self.infer_type(expr)
            if actual_type != needed_type:
                # FIXME: problematic: potential quadratic complexity
                is_varying = IsVaryingMapper(self.kernel)(expr)
                registry = self.codegen_state.ast_builder.target.get_dtype_registry()
                cast = var("("
                           f"{'varying' if is_varying else 'uniform'} "
                           f"{registry.dtype_to_ctype(needed_type)}"
                           ") ")
                return cast(result)

            return result

# }}}


# {{{ type registry

def fill_registry_with_ispc_types(reg, respect_windows, include_bool=True):
    reg.get_or_register_dtype("bool", bool)

    reg.get_or_register_dtype(["int8", "signed char", "char"], np.int8)
    reg.get_or_register_dtype(["uint8", "unsigned char"], np.uint8)
    reg.get_or_register_dtype(["int16", "short", "signed short",
        "signed short int", "short signed int"], np.int16)
    reg.get_or_register_dtype(["uint16", "unsigned short",
        "unsigned short int", "short unsigned int"], np.uint16)
    reg.get_or_register_dtype(["int32", "int", "signed int"], np.int32)
    reg.get_or_register_dtype(["uint32", "unsigned", "unsigned int"], np.uint32)

    reg.get_or_register_dtype(["int64"], np.int64)
    reg.get_or_register_dtype(["uint64"], np.uint64)

    reg.get_or_register_dtype("float", np.float32)
    reg.get_or_register_dtype("double", np.float64)

# }}}


class ISPCTarget(CFamilyTarget):
    """A code generation target for Intel's `ISPC <https://ispc.github.io/>`_
    SPMD programming language, to target Intel's Knight's hardware and modern
    Intel CPUs with wide vector units.
    """

    host_program_name_suffix = ""
    device_program_name_suffix = "_inner"

    def pre_codegen_entrypoint_check(self, kernel, callables_table):
        _gsize, lsize = kernel.get_grid_size_upper_bounds_as_exprs(
                callables_table)
        if len(lsize) > 1:
            for ls_i in lsize[1:]:
                if ls_i != 1:
                    raise LoopyError("local axis %d (0-based) "
                            "has length > 1, which is unsupported "
                            "by ISPC" % ls_i)

    def get_host_ast_builder(self):
        return ISPCASTBuilder(self)

    def get_device_ast_builder(self):
        return ISPCASTBuilder(self)

    # {{{ types

    @memoize_method
    def get_dtype_registry(self):
        from loopy.target.c.compyte.dtypes import DTypeRegistry
        result = DTypeRegistry()
        fill_registry_with_ispc_types(result, respect_windows=False,
                include_bool=True)
        return result

    # }}}


class ISPCASTBuilder(CFamilyASTBuilder):
    # {{{ top-level codegen

    def get_function_declaration(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult, schedule_index: int
            ) -> tuple[Sequence[tuple[str, str]], Generable]:
        name = codegen_result.current_program(codegen_state).name
        kernel = codegen_state.kernel

        assert codegen_state.kernel.linearization is not None
        subkernel_name = cast(
                        "CallKernel",
                        codegen_state.kernel.linearization[schedule_index]
                        ).kernel_name

        from cgen import FunctionDeclaration, Value
        from cgen.ispc import ISPCExport, ISPCTask

        if codegen_state.is_entrypoint:
            # subkernel launches occur only as part of entrypoint kernels for now
            from loopy.schedule.tools import get_subkernel_arg_info
            skai = get_subkernel_arg_info(codegen_state.kernel, subkernel_name)
            passed_names = skai.passed_names
            written_names = skai.written_names
        else:
            passed_names = [arg.name for arg in kernel.args]
            written_names = kernel.get_written_variables()

        arg_decls = [self.arg_to_cgen_declarator(
                            kernel, arg_name,
                            is_written=arg_name in written_names)
                        for arg_name in passed_names]

        if codegen_state.is_generating_device_code:
            result: Declarator = ISPCTask(
                        FunctionDeclaration(
                            Value("void", name),
                            arg_decls))
        else:
            result = ISPCExport(
                    FunctionDeclaration(
                        Value("void", name),
                        arg_decls))

        from loopy.target.c import FunctionDeclarationWrapper
        return [], FunctionDeclarationWrapper(result)

    def get_kernel_call(self, codegen_state: CodeGenerationState,
            subkernel_name: str,
            gsize: tuple[Expression, ...],
            lsize: tuple[Expression, ...]) -> Generable:
        kernel = codegen_state.kernel
        ecm = self.get_expression_to_code_mapper(codegen_state)

        from pymbolic.mapper.stringifier import PREC_NONE
        result = []
        from cgen import Block, Statement as S
        if lsize:
            result.append(
                    S(
                        "assert(programCount == (%s))"
                        % ecm(lsize[0], PREC_NONE)))

        if codegen_state.is_entrypoint:
            # subkernel launches occur only as part of entrypoint kernels for now
            from loopy.schedule.tools import get_subkernel_arg_info
            skai = get_subkernel_arg_info(codegen_state.kernel, subkernel_name)
            passed_names = skai.passed_names
        else:
            passed_names = [arg.name for arg in kernel.args]

        from cgen.ispc import ISPCLaunch
        result.append(
                ISPCLaunch(
                    tuple(ecm(gs_i, PREC_NONE) for gs_i in gsize),
                    "{}({})".format(
                        subkernel_name,
                        ", ".join(passed_names)
                        )))

        return Block(result)

    # }}}

    # {{{ code generation guts

    def get_expression_to_c_expression_mapper(self, codegen_state):
        return ExprToISPCExprMapper(codegen_state)

    def add_vector_access(self, access_expr, index):
        return access_expr[index]

    def emit_barrier(self, synchronization_kind, mem_kind, comment):
        from cgen import Comment, Statement

        assert comment

        if synchronization_kind == "local":
            return Comment("local barrier: %s" % comment)

        elif synchronization_kind == "global":
            return Statement("sync; /* %s */" % comment)

        else:
            raise LoopyError("unknown barrier kind")

    # }}}

    # {{{ declarators

    def get_value_arg_declarator(
            self, name: str, dtype: LoopyType, is_written: bool) -> Declarator:
        from cgen.ispc import ISPCUniform
        return ISPCUniform(super().get_value_arg_declarator(
                name, dtype, is_written))

    def get_array_arg_declarator(
            self, arg: ArrayArg, is_written: bool) -> Declarator:
        # FIXME restrict?
        from cgen.ispc import ISPCUniform, ISPCUniformPointer
        decl: Declarator = ISPCUniform(
                ISPCUniformPointer(self.get_array_base_declarator(arg)))

        if not is_written:
            decl = Const(decl)

        return decl

    def get_temporary_var_declarator(self,
            codegen_state: CodeGenerationState,
            temp_var: TemporaryVariable) -> Declarator:
        temp_var_decl = self.get_array_base_declarator(temp_var)

        shape = temp_var.shape

        assert isinstance(shape, tuple)

        if temp_var.address_space == AddressSpace.PRIVATE:
            # FIXME: This is a pretty coarse way of deciding what
            # private temporaries get duplicated. Refine? (See also
            # above in expr to code mapper)
            _, lsize = codegen_state.kernel.get_grid_size_upper_bounds_as_exprs(
                    codegen_state.callables_table)
            shape = lsize + shape

        if shape:
            from cgen import ArrayOf
            ecm = self.get_expression_to_code_mapper(codegen_state)
            temp_var_decl = ArrayOf(
                    temp_var_decl,
                    ecm(p.flattened_product(shape),
                        prec=PREC_NONE, type_context="i"))

        return temp_var_decl

    # }}}

    # {{{ emit_...

    def emit_assignment(
                self,
                codegen_state: CodeGenerationState,
                insn: Assignment
            ):
        kernel = codegen_state.kernel
        ecm = codegen_state.expression_to_code_mapper

        assignee_var_name, = insn.assignee_var_names()

        lhs_var = codegen_state.kernel.get_var_descriptor(assignee_var_name)
        lhs_dtype = lhs_var.dtype

        if insn.atomicity:
            raise NotImplementedError("atomic ops in ISPC")

        from pymbolic.mapper.stringifier import PREC_NONE

        from loopy.expression import dtype_to_type_context

        rhs_type_context = dtype_to_type_context(kernel.target, lhs_dtype)
        rhs_code = ecm(insn.expression, prec=PREC_NONE,
                    type_context=rhs_type_context,
                    needed_dtype=lhs_dtype)

        lhs = insn.assignee

        # {{{ handle streaming stores

        from loopy.kernel.instruction import UseStreamingStoreTag
        if UseStreamingStoreTag() in insn.tags:
            ary = ecm.find_array(lhs)

            from pymbolic import evaluate

            from loopy.kernel.array import get_access_info
            from loopy.symbolic import simplify_using_aff

            if not isinstance(lhs, p.Subscript):
                raise LoopyError("streaming store must have a subscript as argument")

            from loopy.kernel.data import ArrayArg, TemporaryVariable
            if not isinstance(ary, (ArrayArg, TemporaryVariable)):
                raise LoopyError("array type not supported in ISPC: %s"
                        % type(ary).__name)

            index_tuple = tuple(
                    simplify_using_aff(kernel, idx) for idx in lhs.index_tuple)

            access_info = get_access_info(kernel, ary, index_tuple,
                    lambda expr: cast("int",
                                      evaluate(expr, codegen_state.var_subst_map)),
                    codegen_state.vectorization_info)

            l0_inames = {
                iname for iname in insn.within_inames
                if kernel.inames[iname].tags_of_type(LocalInameTag)}

            if len(access_info.subscripts) != 1:
                raise LoopyError("streaming stores must have a subscript")
            subscript, = access_info.subscripts

            if l0_inames:
                l0_iname, = l0_inames
                coeffs = CoefficientCollector([l0_iname])(subscript)
                if coeffs[p.Variable(l0_iname)] != 1:
                    raise ValueError("coefficient of streaming store index "
                                     "in l.0 variable must be 1")

                subscript = flatten(
                    SubstitutionMapper(make_subst_func({l0_iname: 0}))(subscript))
                del l0_iname

            if access_info.vector_index is not None:
                raise LoopyError("streaming store may not use a short-vector "
                        "data type")

            if (l0_inames
                    and not IsVaryingMapper(codegen_state.kernel)(insn.expression)):
                # rhs is uniform, must be cast to varying in order for streaming_store
                # to perform a vector store.
                registry = codegen_state.ast_builder.target.get_dtype_registry()
                rhs_code = var("(varying "
                           f"{registry.dtype_to_ctype(lhs_dtype)}"
                           f") ({rhs_code})")

            from cgen import Statement
            return Statement(
                    "streaming_store(%s + %s, %s)"
                    % (
                        access_info.array_name,
                        ecm(subscript, PREC_NONE, "i"),
                        rhs_code))

        # }}}

        from cgen import Assign
        return Assign(ecm(lhs, prec=PREC_NONE, type_context=None), rhs_code)

    def emit_sequential_loop(self, codegen_state, iname, iname_dtype,
            lbound, ubound, inner, hints):
        ecm = codegen_state.expression_to_code_mapper

        from cgen import For, InlineInitializer
        from cgen.ispc import ISPCUniform
        from pymbolic.mapper.stringifier import PREC_NONE

        from loopy.target.c import POD

        loop = For(
                InlineInitializer(
                    ISPCUniform(POD(self, iname_dtype, iname)),
                    ecm(lbound, PREC_NONE, "i")),
                ecm(
                    p.Comparison(var(iname), "<=", ubound),
                    PREC_NONE, "i"),
                "++%s" % iname,
                inner)

        if hints:
            return Collection([*list(hints), loop])
        else:
            return loop

    # }}}


# TODO: Generate launch code
# TODO: Vector types (element access: done)

# vim: foldmethod=marker



import numpy as np

import pyopencl as cl  # noqa
from pyopencl.tools import (
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,  # noqa
)

import loopy as lp


1/0  # not ready


def test_laplacian(ctx_factory):
    1/0  # not adapted to new language

    dtype = np.float32
    ctx = ctx_factory()
    order = "C"

    n = 8

    from pymbolic import var
    K_sym = var("K")  # noqa

    field_shape = (K_sym, n, n, n)

    # load:     1+6 fields + 1/N D entry
    # store:    1   fields
    # perform:  N*2*6 + 3*5 flops
    # ratio:   (12*N+15)/8  flops per 4 bytes on bus
    #          ~ 14 FLOPS per 4 bytes at N=8
    #          ~ 525 GFLOPS max on a 150GB/s device at N=8 if done perfectly

    # K - run-time symbolic
    knl = lp.make_kernel(ctx.devices[0],
            "[K] -> {[i,j,k,e,m,o1,o2,o3,gi]: 0<=i,j,k,m,o1,o2,o3<%d and 0<=e<K and 0<=gi<6}" % n,  # noqa
            [
                "CSE: ur(i,j,k) = sum_float32(o1, D[i,o1]*cse(u[e,o1,j,k], urf))",
                "CSE: us(i,j,k) = sum_float32(o2, D[j,o2]*cse(u[e,i,o2,k], usf))",
                "CSE: ut(i,j,k) = sum_float32(o3, D[k,o3]*cse(u[e,i,j,o3], utf))",

                # define function
                "CSE: Gu(i,j,k) = G[0,e,i,j,k]*ur(i,j,k) + G[1,e,i,j,k]*us(i,j,k) + G[2,e,i,j,k]*ut(i,j,k)",  # noqa
                "CSE: Gv(i,j,k) = G[1,e,i,j,k]*ur(i,j,k) + G[3,e,i,j,k]*us(i,j,k) + G[4,e,i,j,k]*ut(i,j,k)",  # noqa
                "CSE: Gw(i,j,k) = G[2,e,i,j,k]*ur(i,j,k) + G[4,e,i,j,k]*us(i,j,k) + G[5,e,i,j,k]*ut(i,j,k)",  # noqa

                "lap[e,i,j,k]  = "
                "  sum_float32(m, D[m,i]*Gu(m,j,k))"
                "+ sum_float32(m, D[m,j]*Gv(i,m,k))"
                "+ sum_float32(m, D[m,k]*Gw(i,j,m))"
                ],
            [
            lp.GlobalArg("u", dtype, shape=field_shape, order=order),
            lp.GlobalArg("lap", dtype, shape=field_shape, order=order),
            lp.GlobalArg("G", dtype, shape=(6, *field_shape), order=order),
            lp.GlobalArg("D", dtype, shape=(n, n), order=order),
            lp.ValueArg("K", np.int32, approximately=1000),
            ],
            name="semlap", assumptions="K>=1")

    # print lp.preprocess_kernel(knl, cse_ok=True)
    # 1/0
    #
    # print knl
    # 1/0
    knl = lp.realize_cse(knl, "urf", np.float32, ["o1"])
    knl = lp.realize_cse(knl, "usf", np.float32, ["o2"])
    knl = lp.realize_cse(knl, "utf", np.float32, ["o3"])

    knl = lp.realize_cse(knl, "Gu", np.float32, ["m", "j", "k"])
    knl = lp.realize_cse(knl, "Gv", np.float32, ["i", "m", "k"])
    knl = lp.realize_cse(knl, "Gw", np.float32, ["i", "j", "m"])

    knl = lp.realize_cse(knl, "ur", np.float32, ["k", "j", "m"])
    knl = lp.realize_cse(knl, "us", np.float32, ["i", "m", "k"])
    knl = lp.realize_cse(knl, "ut", np.float32, ["i", "j", "m"])

    if 0:
        pass
        #seq_knl = lp.add_prefetch(knl, "G", ["gi", "m", "j", "k"], "G[gi,e,m,j,k]", default_tag="l.auto")  # noqa
        # seq_knl = lp.add_prefetch(seq_knl, "D", ["m", "j"], default_tag="l.auto")
        #seq_knl = lp.add_prefetch(seq_knl, "u", ["i", "j", "k"], "u[*,i,j,k]", default_tag="l.auto")  # noqa
    else:
        seq_knl = knl

    knl = lp.split_iname(knl, "e", 16, outer_tag="g.0")  # , slabs=(0, 1))

    knl = lp.add_prefetch(knl, "G", ["gi", "m", "j", "k"], "G[gi,e,m,j,k]", default_tag="l.auto")  # noqa
    knl = lp.add_prefetch(knl, "D", ["m", "j"], default_tag="l.auto")
    #knl = lp.add_prefetch(knl, "u", ["i", "j", "k"], "u[*,i,j,k]", default_tag="l.auto")  # noqa

    # knl = lp.split_iname(knl, "e_inner", 4, inner_tag="ilp")

    # print seq_knl
    # print lp.preprocess_kernel(knl)
    # 1/0

    knl = lp.tag_inames(knl, {"i": "l.0", "j": "l.1"})

    kernel_gen = lp.generate_loop_schedules(knl,
            loop_priority=["m_fetch_G", "i_fetch_u"])
    kernel_gen = lp.check_kernels(kernel_gen, {"K": 1000})

    K = 1000  # noqa
    lp.auto_test_vs_ref(seq_knl, ctx, kernel_gen,
            op_count=K*(n*n*n*n*2*3 + n*n*n*5*3 + n**4 * 2*3)/1e9,
            op_label="GFlops",
            parameters={"K": K}, print_seq_code=True)


def test_laplacian_lmem(ctx_factory):
    1/0  # not adapted to new language

    dtype = np.float32
    ctx = ctx_factory()
    order = "C"

    n = 8

    from pymbolic import var
    K_sym = var("K")  # noqa

    field_shape = (K_sym, n, n, n)

    # K - run-time symbolic
    knl = lp.make_kernel(ctx.devices[0],
            "[K] -> {[i,j,k,e,m,o,gi]: 0<=i,j,k,m,o<%d and 0<=e<K and 0<=gi<6}" % n,
            [
                "CSE: ur(i,j,k) = sum_float32(@o, D[i,o]*u[e,o,j,k])",
                "CSE: us(i,j,k) = sum_float32(@o, D[j,o]*u[e,i,o,k])",
                "CSE: ut(i,j,k) = sum_float32(@o, D[k,o]*u[e,i,j,o])",

                "lap[e,i,j,k]  = "
                "  sum_float32(m, D[m,i]*(G[0,e,m,j,k]*ur(m,j,k) + G[1,e,m,j,k]*us(m,j,k) + G[2,e,m,j,k]*ut(m,j,k)))"  # noqa
                "+ sum_float32(m, D[m,j]*(G[1,e,i,m,k]*ur(i,m,k) + G[3,e,i,m,k]*us(i,m,k) + G[4,e,i,m,k]*ut(i,m,k)))"  # noqa
                "+ sum_float32(m, D[m,k]*(G[2,e,i,j,m]*ur(i,j,m) + G[4,e,i,j,m]*us(i,j,m) + G[5,e,i,j,m]*ut(i,j,m)))"  # noqa
                ],
            [
            lp.GlobalArg("u", dtype, shape=field_shape, order=order),
            lp.GlobalArg("lap", dtype, shape=field_shape, order=order),
            lp.GlobalArg("G", dtype, shape=(6, *field_shape), order=order),
            lp.GlobalArg("D", dtype, shape=(n, n), order=order),
            lp.ValueArg("K", np.int32, approximately=1000),
            ],
            name="semlap", assumptions="K>=1")

    knl = lp.realize_cse(knl, "ur", np.float32, ["k", "j", "m"])
    knl = lp.realize_cse(knl, "us", np.float32, ["i", "m", "k"])
    knl = lp.realize_cse(knl, "ut", np.float32, ["i", "j", "m"])

    if 0:
        seq_knl = lp.add_prefetch(knl, "G", ["gi", "m", "j", "k"], "G[gi,e,m,j,k]",
                default_tag="l.auto")
        seq_knl = lp.add_prefetch(seq_knl, "D", ["m", "j"],
                default_tag="l.auto")
        seq_knl = lp.add_prefetch(seq_knl, "u", ["i", "j", "k"], "u[*,i,j,k]",
                default_tag="l.auto")
    else:
        seq_knl = knl

    knl = lp.split_iname(knl, "e", 16, outer_tag="g.0")  # , slabs=(0, 1))

    knl = lp.add_prefetch(knl, "G", ["gi", "m", "j", "k"], "G[gi,e,m,j,k]",
            default_tag="l.auto")
    knl = lp.add_prefetch(knl, "D", ["m", "j"],
            default_tag="l.auto")
    knl = lp.add_prefetch(knl, "u", ["i", "j", "k"], "u[*,i,j,k]",
            default_tag="l.auto")
    # knl = lp.split_iname(knl, "e_inner", 4, inner_tag="ilp")

    # print seq_knl
    # print lp.preprocess_kernel(knl)
    # 1/0

    knl = lp.tag_inames(knl, {"i": "l.0", "j": "l.1"})

    kernel_gen = lp.generate_loop_schedules(knl)
    kernel_gen = lp.check_kernels(kernel_gen, {"K": 1000})

    K = 1000  # noqa
    lp.auto_test_vs_ref(seq_knl, ctx, kernel_gen,
            op_count=K*(n*n*n*n*2*3 + n*n*n*5*3 + n**4 * 2*3)/1e9,
            op_label="GFlops",
            parameters={"K": K}, print_seq_code=True)


def test_laplacian_lmem_ilp(ctx_factory):
    # This does not lead to practical/runnable code (out of lmem), but it's an
    # excellent stress test for the code generator. :)

    dtype = np.float32
    ctx = ctx_factory()
    order = "C"

    n = 8

    from pymbolic import var
    K_sym = var("K")  # noqa

    field_shape = (K_sym, n, n, n)

    # K - run-time symbolic
    knl = lp.make_kernel(ctx.devices[0],
            "[K] -> {[i,j,k,e,m,o,gi]: 0<=i,j,k,m,o<%d and 0<=e<K }" % n,
            [
                "ur(i,j,k) := sum_float32(@o, D[i,o]*u[e,o,j,k])",
                "us(i,j,k) := sum_float32(@o, D[j,o]*u[e,i,o,k])",
                "ut(i,j,k) := sum_float32(@o, D[k,o]*u[e,i,j,o])",

                "lap[e,i,j,k]  = "
                "  sum_float32(m, D[m,i]*(G[0,e,m,j,k]*ur(m,j,k) + G[1,e,m,j,k]*us(m,j,k) + G[2,e,m,j,k]*ut(m,j,k)))"  # noqa
                "+ sum_float32(m, D[m,j]*(G[1,e,i,m,k]*ur(i,m,k) + G[3,e,i,m,k]*us(i,m,k) + G[4,e,i,m,k]*ut(i,m,k)))"  # noqa
                "+ sum_float32(m, D[m,k]*(G[2,e,i,j,m]*ur(i,j,m) + G[4,e,i,j,m]*us(i,j,m) + G[5,e,i,j,m]*ut(i,j,m)))"  # noqa
                ],
            [
            lp.GlobalArg("u", dtype, shape=field_shape, order=order),
            lp.GlobalArg("lap", dtype, shape=field_shape, order=order),
            lp.GlobalArg("G", dtype, shape=(6, *field_shape), order=order),
            lp.GlobalArg("D", dtype, shape=(n, n), order=order),
            lp.ValueArg("K", np.int32, approximately=1000),
            ],
            name="semlap", assumptions="K>=1")

    # Must act on u first, otherwise stencil becomes crooked and
    # footprint becomes non-convex.

    knl = lp.split_iname(knl, "e", 16, outer_tag="g.0")  # , slabs=(0, 1))
    knl = lp.split_iname(knl, "e_inner", 4, inner_tag="ilp")

    knl = lp.add_prefetch(knl, "u", [1, 2, 3, "e_inner_inner"],
            default_tag="l.auto")

    knl = lp.precompute(knl, "ur", np.float32, [0, 1, 2, "e_inner_inner"],
            default_tag="l.auto")
    knl = lp.precompute(knl, "us", np.float32, [0, 1, 2, "e_inner_inner"],
            default_tag="l.auto")
    knl = lp.precompute(knl, "ut", np.float32, [0, 1, 2, "e_inner_inner"],
            default_tag="l.auto")

    knl = lp.add_prefetch(knl, "G", ["m", "i", "j", "k", "e_inner_inner"],
            default_tag="l.auto")
    knl = lp.add_prefetch(knl, "D", ["m", "j"],
            default_tag="l.auto")

    # print seq_knl
    # 1/0

    knl = lp.tag_inames(knl, {"i": "l.0", "j": "l.1"})

    kernel_gen = lp.generate_loop_schedules(knl)
    kernel_gen = lp.check_kernels(kernel_gen, {"K": 1000})

    for knl in kernel_gen:
        print(lp.generate_code(knl))


def test_advect(ctx_factory):
    1/0  # not ready

    dtype = np.float32
    ctx = ctx_factory()

    order = "C"

    N = 8  # noqa

    from pymbolic import var
    K_sym = var("K")  # noqa

    field_shape = (K_sym, N, N, N)

    # 1. direction-by-direction similarity transform on u
    # 2. invert diagonal
    # 3. transform back (direction-by-direction)

    # K - run-time symbolic

    # A. updated for CSE: notation.
    # B. fixed temp indexing and C ordering
    # load:     3+9 fields + 1/N D entry
    # store:    3   fields
    # perform:  N*2*6 + 3*5 + 3*5 flops
    # ratio:   (12*N+30)/15  flops per 4 bytes on bus
    #          ~ 8.4 FLOPS per 4 bytes at N=8
    #          ~ 300 GFLOPS max on a 150GB/s device at N=8 if done perfectly
    knl = lp.make_kernel(ctx.devices[0],
            "[K] -> {[i,j,k,m,e]: 0<=i,j,k,m<%d AND 0<=e<K}" % N,
            [
                # differentiate u
                "CSE:  ur(i,j,k) = sum_float32(@m, D[i,m]*u[e,m,j,k])",
                "CSE:  us(i,j,k) = sum_float32(@m, D[j,m]*u[e,i,m,k])",
                "CSE:  ut(i,j,k) = sum_float32(@m, D[k,m]*u[e,i,j,m])",

                # differentiate v
                "CSE:  vr(i,j,k) = sum_float32(@m, D[i,m]*v[e,m,j,k])",
                "CSE:  vs(i,j,k) = sum_float32(@m, D[j,m]*v[e,i,m,k])",
                "CSE:  vt(i,j,k) = sum_float32(@m, D[k,m]*v[e,i,j,m])",

                # differentiate w
                "CSE:  wr(i,j,k) = sum_float32(@m, D[i,m]*w[e,m,j,k])",
                "CSE:  ws(i,j,k) = sum_float32(@m, D[j,m]*w[e,i,m,k])",
                "CSE:  wt(i,j,k) = sum_float32(@m, D[k,m]*w[e,i,j,m])",

                # find velocity in (r,s,t) coordinates
                # CSE?
                "CSE: Vr(i,j,k) = G[0,e,i,j,k]*u[e,i,j,k] + G[1,e,i,j,k]*v[e,i,j,k] + G[2,e,i,j,k]*w[e,i,j,k]",  # noqa
                "CSE: Vs(i,j,k) = G[3,e,i,j,k]*u[e,i,j,k] + G[4,e,i,j,k]*v[e,i,j,k] + G[5,e,i,j,k]*w[e,i,j,k]",  # noqa
                "CSE: Vt(i,j,k) = G[6,e,i,j,k]*u[e,i,j,k] + G[7,e,i,j,k]*v[e,i,j,k] + G[8,e,i,j,k]*w[e,i,j,k]",  # noqa

                # form nonlinear term on integration nodes
                "Nu[e,i,j,k] = Vr(i,j,k)*ur(i,j,k)+Vs(i,j,k)*us(i,j,k)+Vt(i,j,k)*ut(i,j,k)",  # noqa
                "Nv[e,i,j,k] = Vr(i,j,k)*vr(i,j,k)+Vs(i,j,k)*vs(i,j,k)+Vt(i,j,k)*vt(i,j,k)",  # noqa
                "Nw[e,i,j,k] = Vr(i,j,k)*wr(i,j,k)+Vs(i,j,k)*ws(i,j,k)+Vt(i,j,k)*wt(i,j,k)",  # noqa
                ],
            [
            lp.GlobalArg("u",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("v",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("w",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("Nu",  dtype, shape=field_shape, order=order),
            lp.GlobalArg("Nv",  dtype, shape=field_shape, order=order),
            lp.GlobalArg("Nw",  dtype, shape=field_shape, order=order),
            lp.GlobalArg("G",   dtype, shape=(9, *field_shape), order=order),
            lp.GlobalArg("D",   dtype, shape=(N, N),  order=order),
            lp.ValueArg("K",  np.int32, approximately=1000),
            ],
            name="sem_advect", assumptions="K>=1")

    print(knl)
    1/0

    seq_knl = knl

    knl = lp.split_iname(knl, "e", 16, outer_tag="g.0")  # , slabs=(0, 1))

    knl = lp.tag_inames(knl, {"i": "l.0", "j": "l.1"})

    kernel_gen = lp.generate_loop_schedules(knl)
    kernel_gen = lp.check_kernels(kernel_gen, {"K": 1000}, kill_level_min=5)

    K = 1000  # noqa
    lp.auto_test_vs_ref(seq_knl, ctx, kernel_gen,
            op_count=0,
            op_label="GFlops",
            parameters={"K": K}, print_seq_code=True,)


def test_advect_dealias(ctx_factory):
    1/0  # not ready

    dtype = np.float32
    ctx = ctx_factory()
    order = "C"

    N = 8  # noqa
    M = 8  # noqa

    from pymbolic import var
    K_sym = var("K")  # noqa

    field_shape = (N, N, N, K_sym)

    # 1. direction-by-direction similarity transform on u
    # 2. invert diagonal
    # 3. transform back (direction-by-direction)

    # K - run-time symbolic
    knl = lp.make_kernel(ctx.devices[0],
            "[K] -> {[i,ip,j,jp,k,kp,m,e]: 0<=i,j,k,m<%d AND 0<=o,ip,jp,kp<%d 0<=e<K}" %M %N  # noqa
            [

                # interpolate u to integration nodes
                "CSE:  u0[i,jp,kp,e] = sum_float32(@o, I[i,o]*u[o,jp,kp,e])",
                "CSE:  u1[i,j,kp,e]  = sum_float32(@o, I[j,o]*u0[i,o,kp,e])",
                "CSE:  Iu[i,j,k,e]   = sum_float32(@o, I[k,o]*u1[i,j,o,e])",

                # differentiate u on integration nodes
                "CSE:  Iur[i,j,k,e]  = sum_float32(@m, D[i,m]*Iu[m,j,k,e])",
                "CSE:  Ius[i,j,k,e]  = sum_float32(@m, D[j,m]*Iu[i,m,k,e])",
                "CSE:  Iut[i,j,k,e]  = sum_float32(@m, D[k,m]*Iu[i,j,m,e])",

                # interpolate v to integration nodes
                "CSE:  v0[i,jp,kp,e] = sum_float32(@o, I[i,o]*v[o,jp,kp,e])",
                "CSE:  v1[i,j,kp,e]  = sum_float32(@o, I[j,o]*v0[i,o,kp,e])",
                "CSE:  Iv[i,j,k,e]   = sum_float32(@o, I[k,o]*v1[i,j,o,e])",

                # differentiate v on integration nodes
                "CSE:  Ivr[i,j,k,e]  = sum_float32(@m, D[i,m]*Iv[m,j,k,e])",
                "CSE:  Ivs[i,j,k,e]  = sum_float32(@m, D[j,m]*Iv[i,m,k,e])",
                "CSE:  Ivt[i,j,k,e]  = sum_float32(@m, D[k,m]*Iv[i,j,m,e])",

                # interpolate w to integration nodes
                "CSE:  w0[i,jp,kp,e] = sum_float32(@o, I[i,o]*w[o,jp,kp,e])",
                "CSE:  w1[i,j,kp,e]  = sum_float32(@o, I[j,o]*w0[i,o,kp,e])",
                "CSE:  Iw[i,j,k,e]   = sum_float32(@o, I[k,o]*w1[i,j,o,e])",

                # differentiate v on integration nodes
                "CSE:  Iwr[i,j,k,e]  = sum_float32(@m, D[i,m]*Iw[m,j,k,e])",
                "CSE:  Iws[i,j,k,e]  = sum_float32(@m, D[j,m]*Iw[i,m,k,e])",
                "CSE:  Iwt[i,j,k,e]  = sum_float32(@m, D[k,m]*Iw[i,j,m,e])",

                # find velocity in (r,s,t) coordinates
                # QUESTION: should I use CSE here ?
                "CSE: Vr[i,j,k,e] = G[i,j,k,0,e]*Iu[i,j,k,e] + G[i,j,k,1,e]*Iv[i,j,k,e] + G[i,j,k,2,e]*Iw[i,j,k,e]",  # noqa
                "CSE: Vs[i,j,k,e] = G[i,j,k,3,e]*Iu[i,j,k,e] + G[i,j,k,4,e]*Iv[i,j,k,e] + G[i,j,k,5,e]*Iw[i,j,k,e]",  # noqa
                "CSE: Vt[i,j,k,e] = G[i,j,k,6,e]*Iu[i,j,k,e] + G[i,j,k,7,e]*Iv[i,j,k,e] + G[i,j,k,8,e]*Iw[i,j,k,e]",  # noqa

                # form nonlinear term on integration nodes
                # QUESTION: should I use CSE here ?
                "<SE: Nu[i,j,k,e] = Vr[i,j,k,e]*Iur[i,j,k,e]+Vs[i,j,k,e]*Ius[i,j,k,e]+Vt[i,j,k,e]*Iut[i,j,k,e]",  # noqa
                "<SE: Nv[i,j,k,e] = Vr[i,j,k,e]*Ivr[i,j,k,e]+Vs[i,j,k,e]*Ivs[i,j,k,e]+Vt[i,j,k,e]*Ivt[i,j,k,e]",  # noqa
                "<SE: Nw[i,j,k,e] = Vr[i,j,k,e]*Iwr[i,j,k,e]+Vs[i,j,k,e]*Iws[i,j,k,e]+Vt[i,j,k,e]*Iwt[i,j,k,e]",  # noqa

                # L2 project Nu back to Lagrange basis
                "CSE: Nu2[ip,j,k,e]   = sum_float32(@m, V[ip,m]*Nu[m,j,k,e])",
                "CSE: Nu1[ip,jp,k,e]  = sum_float32(@m, V[jp,m]*Nu2[ip,m,k,e])",
                "INu[ip,jp,kp,e] = sum_float32(@m, V[kp,m]*Nu1[ip,jp,m,e])",

                # L2 project Nv back to Lagrange basis
                "CSE: Nv2[ip,j,k,e]   = sum_float32(@m, V[ip,m]*Nv[m,j,k,e])",
                "CSE: Nv1[ip,jp,k,e]  = sum_float32(@m, V[jp,m]*Nv2[ip,m,k,e])",
                "INv[ip,jp,kp,e] = sum_float32(@m, V[kp,m]*Nv1[ip,jp,m,e])",

                # L2 project Nw back to Lagrange basis
                "CSE: Nw2[ip,j,k,e]   = sum_float32(@m, V[ip,m]*Nw[m,j,k,e])",
                "CSE: Nw1[ip,jp,k,e]  = sum_float32(@m, V[jp,m]*Nw2[ip,m,k,e])",
                "INw[ip,jp,kp,e] = sum_float32(@m, V[kp,m]*Nw1[ip,jp,m,e])",

                ],
            [
            lp.GlobalArg("u",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("v",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("w",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("INu",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("INv",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("INw",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("D",   dtype, shape=(M, M), order=order),
            lp.GlobalArg("I",   dtype, shape=(M, N), order=order),
            lp.GlobalArg("V",   dtype, shape=(N, M), order=order),
            lp.ValueArg("K",  np.int32, approximately=1000),
            ],
            name="sem_advect", assumptions="K>=1")

    print(knl)
    1/0

    knl = lp.split_iname(knl, "e", 16, outer_tag="g.0")  # , slabs=(0, 1))

    knl = lp.tag_inames(knl, {"i": "l.0", "j": "l.1"})

    print(knl)
    # 1/0

    kernel_gen = lp.generate_loop_schedules(knl)
    kernel_gen = lp.check_kernels(kernel_gen, {"K": 1000}, kill_level_min=5)

    K = 1000  # noqa
    lp.auto_test_vs_ref(knl, ctx, kernel_gen,
            op_count=0,
            op_label="GFlops",
            parameters={"K": K}, print_seq_code=True,)


def test_interp_diff(ctx_factory):
    1/0  # not ready
    dtype = np.float32
    ctx = ctx_factory()
    order = "C"

    N = 8  # noqa
    M = 8  # noqa

    from pymbolic import var
    K_sym = var("K")  # noqa

    field_shape = (N, N, N, K_sym)
    interim_field_shape = (M, M, M, K_sym)

    # 1. direction-by-direction similarity transform on u
    # 2. invert diagonal
    # 3. transform back (direction-by-direction)

    # K - run-time symbolic
    knl = lp.make_kernel(ctx.devices[0],
            "[K] -> {[i,ip,j,jp,k,kp,e]: 0<=i,j,k<%d AND 0<=ip,jp,kp<%d 0<=e<K}" %M %N  # noqa
            [
                "[|i,jp,kp] <float32>  u1[i ,jp,kp,e] = sum_float32(ip, I[i,ip]*u [ip,jp,kp,e])",  # noqa
                "[|i,j ,kp] <float32>  u2[i ,j ,kp,e] = sum_float32(jp, I[j,jp]*u1[i ,jp,kp,e])",  # noqa
                "[|i,j ,k ] <float32>  u3[i ,j ,k ,e] = sum_float32(kp, I[k,kp]*u2[i ,j ,kp,e])",  # noqa
                "[|i,j ,k ] <float32>  Pu[i ,j ,k ,e] = P[i,j,k,e]*u3[i,j,k,e]",
                "[|i,j ,kp] <float32> Pu3[i ,j ,kp,e] = sum_float32(k, V[kp,k]*Pu[i ,j , k,e])",  # noqa
                "[|i,jp,kp] <float32> Pu2[i ,jp,kp,e] = sum_float32(j, V[jp,j]*Pu[i ,j ,kp,e])",  # noqa
                "Pu[ip,jp,kp,e] = sum_float32(i, V[ip,i]*Pu[i ,jp,kp,e])",
                ],
            [
            lp.GlobalArg("u",   dtype, shape=field_shape, order=order),
            lp.GlobalArg("P",   dtype, shape=interim_field_shape, order=order),
            lp.GlobalArg("I",   dtype, shape=(M, N), order=order),
            lp.GlobalArg("V",   dtype, shape=(N, M), order=order),
            lp.GlobalArg("Pu",  dtype, shape=field_shape, order=order),
            lp.ValueArg("K",  np.int32, approximately=1000),
            ],
            name="sem_lap_precon", assumptions="K>=1")

    print(knl)
    1/0

    knl = lp.split_iname(knl, "e", 16, outer_tag="g.0")  # , slabs=(0, 1))

    knl = lp.tag_inames(knl, {"i": "l.0", "j": "l.1"})

    print(knl)
    # 1/0

    kernel_gen = lp.generate_loop_schedules(knl)
    kernel_gen = lp.check_kernels(kernel_gen, {"K": 1000}, kill_level_min=5)

    K = 1000  # noqa
    lp.auto_test_vs_ref(knl, ctx, kernel_gen,
            op_count=0,
            op_label="GFlops",
            parameters={"K": K}, print_seq_code=True,)


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from py.test.cmdline import main
        main([__file__])



from __future__ import annotations


__copyright__ = "Copyright (C) 2016 Matt Wala"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
from functools import cached_property

from constantdict import constantdict

from pytools import Record, memoize_method

from loopy.diagnostic import LoopyError
from loopy.kernel.data import AddressSpace, Iname
from loopy.schedule import (
    Barrier,
    CallKernel,
    EnterLoop,
    LeaveLoop,
    ReturnFromKernel,
    RunInstruction,
)
from loopy.schedule.tools import get_block_boundaries
from loopy.types import auto


logger = logging.getLogger(__name__)


__doc__ = """
.. currentmodule:: loopy

.. autofunction:: save_and_reload_temporaries
"""


# {{{ liveness analysis

class LivenessResult(dict):

    class InstructionResult(Record):
        __slots__ = ["live_in", "live_out"]

    @classmethod
    def make_empty(cls, nscheditems):
        return cls((idx, cls.InstructionResult(live_in=set(), live_out=set()))
                   for idx in range(nscheditems))


class LivenessAnalysis:

    def __init__(self, kernel):
        self.kernel = kernel
        self.schedule = kernel.linearization

    @memoize_method
    def get_successor_relation(self):
        successors = {}
        block_bounds = get_block_boundaries(self.kernel.linearization)

        for idx, (item, next_item) in enumerate(zip(
                reversed(self.schedule),
                reversed([*self.schedule, None]))):
            sched_idx = len(self.schedule) - idx - 1

            # Look at next_item
            if next_item is None:
                after = set()
            elif isinstance(next_item, EnterLoop):
                # Account for empty loop
                loop_end = block_bounds[sched_idx + 1]
                after = successors[loop_end] | {sched_idx + 1}
            elif isinstance(next_item, (LeaveLoop, RunInstruction,
                    CallKernel, ReturnFromKernel, Barrier)):
                after = {sched_idx + 1}
            else:
                raise LoopyError("unexpected type of schedule item: {ty}"
                    .format(ty=type(next_item).__name__))

            # Look at item
            if isinstance(item, LeaveLoop):
                # Account for loop
                loop_begin = block_bounds[sched_idx]
                after |= {loop_begin}
            elif not isinstance(item, (EnterLoop, RunInstruction,
                    CallKernel, ReturnFromKernel, Barrier)):
                raise LoopyError("unexpected type of schedule item: {ty}"
                    .format(ty=type(item).__name__))

            successors[sched_idx] = after

        return successors

    def get_gen_and_kill_sets(self):
        gen = {idx: set() for idx in range(len(self.schedule))}
        kill = {idx: set() for idx in range(len(self.schedule))}

        for sched_idx, sched_item in enumerate(self.schedule):
            if not isinstance(sched_item, RunInstruction):
                continue
            insn = self.kernel.id_to_insn[sched_item.insn_id]
            for var in insn.assignee_var_names():
                if var not in self.kernel.temporary_variables:
                    continue
                if not insn.predicates:
                    # Fully kills the liveness only when unconditional.
                    kill[sched_idx].add(var)
                if len(self.kernel.temporary_variables[var].shape) > 0:
                    # For an array variable, all definitions generate a use as
                    # well, because the write could be a partial write,
                    # necessitating a reload of whatever is not written.
                    #
                    # We don't currently check if the write is a partial write
                    # or a full write. Instead, we analyze the access
                    # footprint later on to determine how much to reload/save.
                    gen[sched_idx].add(var)
            for var in insn.read_dependency_names():
                if var not in self.kernel.temporary_variables:
                    continue
                gen[sched_idx].add(var)

        return gen, kill

    @memoize_method
    def liveness(self):
        logger.info("running liveness analysis")
        successors = self.get_successor_relation()
        gen, kill = self.get_gen_and_kill_sets()

        # Fixed point iteration for liveness analysis
        lr = LivenessResult.make_empty(len(self.schedule))

        prev_lr = None

        while prev_lr != lr:
            from copy import deepcopy
            prev_lr = deepcopy(lr)
            for idx in range(len(self.schedule) - 1, -1, -1):
                for succ in successors[idx]:
                    lr[idx].live_out.update(lr[succ].live_in)
                lr[idx].live_in = gen[idx] | (lr[idx].live_out - kill[idx])

        logger.info("done running liveness analysis")

        return lr

    def print_liveness(self):
        print(75 * "-")
        print("LIVE IN:")
        for sched_idx in range(len(self.schedule)):
            print("{item}: {{{vars}}}".format(
                item=sched_idx,
                vars=", ".join(sorted(self[sched_idx].live_in))))
        print(75 * "-")
        print("LIVE OUT:")
        for sched_idx in range(len(self.schedule)):
            print("{item}: {{{vars}}}".format(
                item=sched_idx,
                vars=", ".join(sorted(self[sched_idx].live_out))))
        print(75 * "-")

    def __getitem__(self, sched_idx):
        """
        :arg insn: An instruction name or instance of
            :class:`loopy.instruction.InstructionBase`

        :returns: A :class:`LivenessResult` associated with `insn`
        """
        return self.liveness()[sched_idx]

# }}}


# {{{ save and reload implementation

class TemporarySaver:

    class PromotedTemporary(Record):
        """
        .. attribute:: name

            The name of the new temporary.

        .. attribute:: orig_temporary_name

            The name of original temporary variable object.

        .. attribute:: hw_dims

            A list of expressions, to be added in front of the shape
            of the promoted temporary value, corresponding to
            hardware dimensions

        .. attribute:: hw_tags

            The tags for the inames associated with hw_dims

        .. attribute:: non_hw_dims

            A list of expressions, to be added in front of the shape
            of the promoted temporary value, corresponding to
            non-hardware dimensions
        """

        __slots__ = """
                name
                orig_temporary_name
                hw_dims
                hw_tags
                non_hw_dims""".split()

        def as_kernel_temporary(self, kernel):
            temporary = kernel.temporary_variables[self.orig_temporary_name]
            from loopy.kernel.data import TemporaryVariable
            return TemporaryVariable(
                name=self.name,
                dtype=temporary.dtype,
                address_space=AddressSpace.GLOBAL,
                shape=self.new_shape)

        @property
        def new_shape(self):
            return self.hw_dims + self.non_hw_dims

    def __init__(self, kernel, callables_table):
        self.kernel = kernel
        self.callables_table = callables_table
        self.var_name_gen = kernel.get_var_name_generator()
        self.insn_name_gen = kernel.get_instruction_id_generator()

        # These fields keep track of updates to the kernel.
        from collections import defaultdict
        self.insns_to_insert = []
        self.insns_to_update = {}
        self.updated_iname_objs = constantdict()
        self.updated_temporary_variables = {}

        # temporary name -> save or reload insn ids
        self.temporary_to_save_ids = defaultdict(set)
        self.temporary_to_reload_ids = defaultdict(set)
        self.subkernel_to_newly_added_insn_ids = defaultdict(set)

        # Maps names of base_storage to the name of the temporary
        # representative chosen for saves/reloads
        self.base_storage_to_representative = {}

        import islpy as isl

        from loopy.kernel.data import ValueArg
        self.new_subdomain = (
                isl.BasicSet.universe(
                    isl.Space.create_from_names(
                        isl.DEFAULT_CONTEXT,
                        set=[],
                        params={
                            arg.name for arg in kernel.args
                            if isinstance(arg, ValueArg)})))

    def find_accessing_instructions_in_subkernel(self, temporary, subkernel):
        # Find all accessing instructions in the subkernel. If base_storage is
        # present, this includes instructions that access aliasing memory.

        aliasing_names = {temporary}
        base_storage = self.kernel.temporary_variables[temporary].base_storage

        if base_storage is not None:
            aliasing_names |= self.base_storage_to_temporary_map[base_storage]

        from loopy.kernel.tools import get_subkernel_to_insn_id_map
        accessing_insns_in_subkernel = set()
        subkernel_insns = get_subkernel_to_insn_id_map(self.kernel)[subkernel]

        for name in aliasing_names:
            try:
                accessing_insns_in_subkernel |= (
                        self.kernel.reader_map()[name] & subkernel_insns)
            except KeyError:
                pass

            try:
                accessing_insns_in_subkernel |= (
                        self.kernel.writer_map()[name] & subkernel_insns)
            except KeyError:
                pass

        return frozenset(accessing_insns_in_subkernel)

    @cached_property
    def base_storage_to_temporary_map(self):
        from collections import defaultdict

        result = defaultdict(set)

        for temporary in self.kernel.temporary_variables.values():
            if temporary.base_storage is None:
                continue
            result[temporary.base_storage].add(temporary.name)

        return result

    @cached_property
    def subkernel_to_slice_indices(self):
        result = {}

        for sched_item_idx, sched_item in enumerate(self.kernel.linearization):
            if isinstance(sched_item, CallKernel):
                start_idx = sched_item_idx
            elif isinstance(sched_item, ReturnFromKernel):
                result[sched_item.kernel_name] = (start_idx, 1 + sched_item_idx)

        return result

    @cached_property
    def subkernel_to_surrounding_inames(self):
        current_outer_inames = set()
        within_subkernel = False
        result = {}

        for sched_item in self.kernel.linearization:
            if isinstance(sched_item, CallKernel):
                within_subkernel = True
                result[sched_item.kernel_name] = frozenset(current_outer_inames)
            elif isinstance(sched_item, ReturnFromKernel):
                within_subkernel = False
            elif isinstance(sched_item, EnterLoop):
                if not within_subkernel:
                    current_outer_inames.add(sched_item.iname)
            elif isinstance(sched_item, LeaveLoop):
                current_outer_inames.discard(sched_item.iname)

        return result

    @memoize_method
    def get_enclosing_global_barrier_pair(self, subkernel):
        subkernel_start, subkernel_end = (
            self.subkernel_to_slice_indices[subkernel])

        def is_global_barrier(item):
            return isinstance(item, Barrier) and \
                item.synchronization_kind == "global"

        try:
            pre_barrier = next(item for item in
                self.kernel.linearization[subkernel_start::-1]
                if is_global_barrier(item)).originating_insn_id
        except StopIteration:
            pre_barrier = None

        try:
            post_barrier = next(item for item in
                self.kernel.linearization[subkernel_end:]
                if is_global_barrier(item)).originating_insn_id
        except StopIteration:
            post_barrier = None

        return (pre_barrier, post_barrier)

    def get_hw_axis_sizes_and_tags_for_save_slot(self, temporary):
        """
        This is used for determining the amount of global storage needed for saving
        and restoring the temporary across kernel calls, due to hardware
        parallel inames (the inferred axes get prefixed to the number of
        dimensions in the temporary).

        In the case of local temporaries, inames that are tagged
        hw-local do not contribute to the global storage shape.
        """
        accessor_insn_ids = frozenset(
            self.kernel.reader_map()[temporary.name]
            | self.kernel.writer_map()[temporary.name])

        group_tags = None
        local_tags = None

        def _sortedtags(tags):
            return sorted(tags, key=lambda tag: tag.axis)

        for insn_id in accessor_insn_ids:
            insn = self.kernel.id_to_insn[insn_id]

            my_group_tags = []
            my_local_tags = []
            group_tags_originating_insn_id = None

            for iname in insn.within_inames:
                tags = self.kernel.iname_tags(iname)

                if not tags:
                    continue

                from loopy.kernel.data import (
                    ConcurrentTag,
                    GroupInameTag,
                    LocalInameTag,
                    filter_iname_tags_by_type,
                )

                if filter_iname_tags_by_type(tags, GroupInameTag):
                    tag, = filter_iname_tags_by_type(tags, GroupInameTag, 1)
                    my_group_tags.append(tag)
                elif filter_iname_tags_by_type(tags, LocalInameTag):
                    tag, = filter_iname_tags_by_type(tags, LocalInameTag, 1)
                    my_local_tags.append(tag)
                elif filter_iname_tags_by_type(tags, ConcurrentTag):
                    raise LoopyError(
                        "iname '%s' is tagged with '%s' - only "
                        "group and local tags are supported for "
                        "auto save/reload of temporaries" %
                        (iname, tags))

            if group_tags is None:
                group_tags = _sortedtags(my_group_tags)
                local_tags = _sortedtags(my_local_tags)
                group_tags_originating_insn_id = insn_id

            if (
                    group_tags != _sortedtags(my_group_tags)
                    or local_tags != _sortedtags(my_local_tags)):
                raise LoopyError(
                    "inconsistent parallel tags across instructions that access "
                    "'%s' (specifically, instruction '%s' has tags '%s' but "
                    "instruction '%s' has tags '%s')"
                    % (temporary.name,
                       group_tags_originating_insn_id, group_tags + local_tags,
                       insn_id, my_group_tags + my_local_tags))

        if group_tags is None:
            assert local_tags is None
            return (), ()

        group_sizes, local_sizes = (
            self.kernel.get_grid_sizes_for_insn_ids_as_exprs(accessor_insn_ids,
                self.callables_table))

        if temporary.address_space == AddressSpace.LOCAL:
            # Elide local axes in the save slot for local temporaries.
            del local_tags[:]
            local_sizes = ()

        # We set hw_dims to be arranged according to the order:
        #    g.0 < g.1 < ... < l.0 < l.1 < ...
        return (group_sizes + local_sizes), tuple(group_tags + local_tags)

    @memoize_method
    def auto_promote_temporary(self, temporary_name):
        temporary = self.kernel.temporary_variables[temporary_name]

        if temporary.address_space == AddressSpace.GLOBAL:
            # Nothing to be done for global temporaries (I hope)
            return None

        if temporary.initializer is not None:
            # Temporaries with initializers do not need saving/reloading - the
            # code generation takes care of emitting the initializers.
            assert temporary.read_only
            return None

        base_storage_conflict = (
            self.base_storage_to_representative.get(
                temporary.base_storage, temporary) is not temporary)

        if base_storage_conflict:
            raise NotImplementedError(
                "tried to save/reload multiple temporaries with the "
                "same base_storage; this is currently not supported")

        hw_dims, hw_tags = self.get_hw_axis_sizes_and_tags_for_save_slot(temporary)
        non_hw_dims = temporary.shape

        if len(non_hw_dims) == 0 and len(hw_dims) == 0:
            # Scalar not in hardware: ensure at least one dimension.
            non_hw_dims = (1,)

        backing_temporary = self.PromotedTemporary(
            name=self.var_name_gen(temporary.name + "_save_slot"),
            orig_temporary_name=temporary.name,
            hw_dims=hw_dims,
            hw_tags=hw_tags,
            non_hw_dims=non_hw_dims)

        if temporary.base_storage is not None:
            self.base_storage_to_representative[temporary.base_storage] = (
                    backing_temporary)

        return backing_temporary

    def save_or_reload_impl(self, temporary, subkernel, mode,
                             promoted_temporary=auto):
        assert mode in ("save", "reload")

        if promoted_temporary is auto:
            promoted_temporary = self.auto_promote_temporary(temporary)

        if promoted_temporary is None:
            return

        new_subdomain, hw_inames, dim_inames, iname_objs = (
            self.augment_domain_for_save_or_reload(
                self.new_subdomain, promoted_temporary, mode, subkernel))

        self.new_subdomain = new_subdomain

        save_or_load_insn_id = self.insn_name_gen(
            f"{temporary}.{mode}")

        def add_subscript_if_subscript_nonempty(agg, subscript=()):
            from pymbolic.primitives import Subscript, Variable
            if len(subscript) == 0:
                return Variable(agg)
            else:
                return Subscript(
                    Variable(agg),
                    tuple(map(Variable, subscript)))

        orig_temporary = (
            self.kernel.temporary_variables[
                promoted_temporary.orig_temporary_name])
        dim_inames_trunc = dim_inames[:len(orig_temporary.shape)]

        args = (
            add_subscript_if_subscript_nonempty(
                temporary, subscript=dim_inames_trunc),
            add_subscript_if_subscript_nonempty(
                promoted_temporary.name, subscript=hw_inames + dim_inames))

        if mode == "save":
            args = reversed(args)

        accessing_insns_in_subkernel = self.find_accessing_instructions_in_subkernel(
                temporary, subkernel)

        if mode == "save":
            depends_on = accessing_insns_in_subkernel
            update_deps = frozenset()
        elif mode == "reload":
            depends_on = frozenset()
            update_deps = accessing_insns_in_subkernel
        else:
            raise AssertionError()

        pre_barrier, post_barrier = self.get_enclosing_global_barrier_pair(subkernel)

        if pre_barrier is not None:
            depends_on |= {pre_barrier}

        if post_barrier is not None:
            update_deps |= {post_barrier}

        # Create the load / store instruction.
        from loopy.kernel.data import Assignment
        save_or_load_insn = Assignment(
            *args,
            id=save_or_load_insn_id,
            within_inames=(
                self.subkernel_to_surrounding_inames[subkernel]
                | frozenset(hw_inames + dim_inames)),
            within_inames_is_final=True,
            depends_on=depends_on)

        if mode == "save":
            self.temporary_to_save_ids[temporary].add(save_or_load_insn_id)
        else:
            self.temporary_to_reload_ids[temporary].add(save_or_load_insn_id)

        self.subkernel_to_newly_added_insn_ids[subkernel].add(save_or_load_insn_id)

        self.insns_to_insert.append(save_or_load_insn)

        for insn_id in update_deps:
            insn = self.insns_to_update.get(insn_id, self.kernel.id_to_insn[insn_id])
            self.insns_to_update[insn_id] = insn.copy(
                depends_on=insn.depends_on | frozenset([save_or_load_insn_id]))

        self.updated_temporary_variables[promoted_temporary.name] = (
            promoted_temporary.as_kernel_temporary(self.kernel))

        self.updated_iname_objs = self.updated_iname_objs.update(iname_objs)

    @memoize_method
    def finish(self):
        new_instructions = []

        insns_to_insert = {insn.id: insn for insn in self.insns_to_insert}

        for orig_insn in self.kernel.instructions:
            if orig_insn.id in self.insns_to_update:
                new_instructions.append(self.insns_to_update[orig_insn.id])
            else:
                new_instructions.append(orig_insn)
        new_instructions.extend(
            sorted(insns_to_insert.values(), key=lambda insn: insn.id))

        self.updated_iname_objs = self.updated_iname_objs.update(self.kernel.inames)
        self.updated_temporary_variables.update(self.kernel.temporary_variables)

        new_domains = list(self.kernel.domains)
        import islpy as isl
        if self.new_subdomain.dim(isl.dim_type.set) > 0:
            new_domains.append(self.new_subdomain)

        kernel = self.kernel.copy(
            domains=new_domains,
            instructions=new_instructions,
            inames=self.updated_iname_objs,
            temporary_variables=self.updated_temporary_variables,
            overridden_get_grid_sizes_for_insn_ids=None)

        # Add nosync directives to any saves or reloads that were added with a
        # potential dependency chain.
        from loopy.kernel.tools import get_subkernels
        for subkernel in get_subkernels(kernel):
            relevant_insns = self.subkernel_to_newly_added_insn_ids[subkernel]

            from itertools import product

            from loopy.transform.instruction import add_nosync
            for temporary in self.temporary_to_reload_ids:
                for source, sink in product(
                        relevant_insns & self.temporary_to_reload_ids[temporary],
                        relevant_insns & self.temporary_to_save_ids[temporary]):
                    kernel = add_nosync(kernel, "global", source, sink)

        from loopy.kernel.tools import assign_automatic_axes
        return assign_automatic_axes(kernel, self.callables_table)

    def save(self, temporary, subkernel):
        self.save_or_reload_impl(temporary, subkernel, "save")

    def reload(self, temporary, subkernel):
        self.save_or_reload_impl(temporary, subkernel, "reload")

    def augment_domain_for_save_or_reload(self,
            domain, promoted_temporary, mode, subkernel):
        """
        Add new axes to the domain corresponding to the dimensions of
        `promoted_temporary`. These axes will be used in the save/
        reload stage. These get prefixed onto the already existing axes.
        """
        assert mode in ("save", "reload")
        import islpy as isl

        orig_temporary = (
                self.kernel.temporary_variables[
                    promoted_temporary.orig_temporary_name])
        orig_dim = domain.dim(isl.dim_type.set)

        # Tags for newly added inames
        iname_objs = {}

        from loopy.symbolic import aff_from_expr

        # FIXME: Restrict size of new inames to access footprint.

        # Add dimension-dependent inames.
        dim_inames = []
        domain = domain.add_dims(isl.dim_type.set,
                            len(promoted_temporary.non_hw_dims)
                            + len(promoted_temporary.hw_dims))

        for dim_idx, dim_size in enumerate(promoted_temporary.non_hw_dims):
            new_iname = self.insn_name_gen("{name}_{mode}_axis_{dim}_{sk}".
                format(name=orig_temporary.name,
                       mode=mode,
                       dim=dim_idx,
                       sk=subkernel))
            domain = domain.set_dim_name(
                isl.dim_type.set, orig_dim + dim_idx, new_iname)

            if orig_temporary.address_space == AddressSpace.LOCAL:
                # If the temporary has local scope, then loads / stores can
                # be done in parallel.
                from loopy.kernel.data import AutoFitLocalInameTag
                iname_objs[new_iname] = Iname(
                        new_iname, tags=frozenset([AutoFitLocalInameTag()]))

            dim_inames.append(new_iname)

            # Add size information.
            aff = isl.affs_from_space(domain.space)
            domain &= aff[0].le_set(aff[new_iname])
            domain &= aff[new_iname].lt_set(aff_from_expr(domain.space, dim_size))

        dim_offset = orig_dim + len(promoted_temporary.non_hw_dims)

        hw_inames = []
        # Add hardware dims.
        for hw_iname_idx, (hw_tag, dim) in enumerate(
                zip(promoted_temporary.hw_tags, promoted_temporary.hw_dims)):
            new_iname = self.insn_name_gen("{name}_{mode}_hw_dim_{dim}_{sk}".
                format(name=orig_temporary.name,
                       mode=mode,
                       dim=hw_iname_idx,
                       sk=subkernel))
            domain = domain.set_dim_name(
                isl.dim_type.set, dim_offset + hw_iname_idx, new_iname)

            aff = isl.affs_from_space(domain.space)
            domain = (domain
                &
                aff[0].le_set(aff[new_iname])
                &
                aff[new_iname].lt_set(aff_from_expr(domain.space, dim)))

            self.updated_iname_objs = self.updated_iname_objs.set(new_iname,
                    Iname(name=new_iname, tags=frozenset([hw_tag])))
            hw_inames.append(new_iname)

        # The operations on the domain above return a Set object, but the
        # underlying domain should be expressible as a single BasicSet.
        domain_list = domain.get_basic_set_list()
        assert domain_list.n_basic_set() == 1
        domain = domain_list.get_basic_set(0)
        return domain, hw_inames, dim_inames, iname_objs

# }}}


# {{{ auto save and reload across kernel calls

def save_and_reload_temporaries(program, entrypoint=None):
    """
    Add instructions to save and reload temporary variables that are live
    across kernel calls.

    The basic code transformation turns schedule segments::

        t = <...>
        <return followed by call>
        <...> = t

    into this code::

        t = <...>
        t_save_slot = t
        <return followed by call>
        t = t_save_slot
        <...> = t

    where `t_save_slot` is a newly-created global temporary variable.

    :returns: The resulting kernel
    """
    if entrypoint is None:
        if len(program.entrypoints) != 1:
            raise LoopyError("Missing argument 'entrypoint'.")
        entrypoint = next(iter(program.entrypoints))

    knl = program[entrypoint]

    from loopy.preprocess import preprocess_program

    if not knl.linearization:
        program = preprocess_program(program)
        from loopy.schedule import get_one_linearized_kernel
        knl = get_one_linearized_kernel(program[entrypoint],
                program.callables_table)

    assert knl.linearization is not None

    liveness = LivenessAnalysis(knl)
    saver = TemporarySaver(knl, program.callables_table)

    from loopy.schedule.tools import (
        temporaries_read_in_subkernel,
        temporaries_written_in_subkernel,
    )

    for sched_idx, sched_item in enumerate(knl.linearization):

        if isinstance(sched_item, CallKernel):
            # Any written temporary that is live-out needs to be read into
            # memory because of the potential for partial writes.
            if sched_idx == 0:
                # Kernel entry: nothing live
                interesting_temporaries = set()
            else:
                subkernel = sched_item.kernel_name
                interesting_temporaries = (
                    temporaries_read_in_subkernel(knl, subkernel)
                    | temporaries_written_in_subkernel(knl,
                                                       subkernel))

            for temporary in liveness[sched_idx].live_out & interesting_temporaries:
                logger.info("reloading {} at entry of {}"
                        .format(temporary, sched_item.kernel_name))
                saver.reload(temporary, sched_item.kernel_name)

        elif isinstance(sched_item, ReturnFromKernel):
            if sched_idx == len(knl.linearization) - 1:
                # Kernel exit: nothing live
                interesting_temporaries = set()
            else:
                subkernel = sched_item.kernel_name
                interesting_temporaries = (
                    temporaries_written_in_subkernel(knl, subkernel))

            for temporary in liveness[sched_idx].live_in & interesting_temporaries:
                logger.info("saving {} before return of {}"
                        .format(temporary, sched_item.kernel_name))
                saver.save(temporary, sched_item.kernel_name)

    return program.with_kernel(saver.finish())

# }}}


# vim: foldmethod=marker



"""
.. currentmodule:: loopy

.. autoclass:: TargetBase
.. autoclass:: ASTBuilderBase
.. autoclass:: CFamilyTarget
.. autoclass:: CTarget
.. autoclass:: ExecutableCTarget
.. autoclass:: CudaTarget
.. autoclass:: OpenCLTarget
.. autoclass:: PyOpenCLTarget
.. autoclass:: ISPCTarget
"""

from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Sequence,
    TypeVar,
)


if TYPE_CHECKING:
    from loopy.codegen import CodeGenerationState
    from loopy.codegen.result import CodeGenerationResult
    from loopy.target.execution import ExecutorBase
    from loopy.translation_unit import FunctionIdT, TranslationUnit
    from loopy.typing import Expression


ASTType = TypeVar("ASTType")


class TargetBase:
    """Base class for all targets, i.e. different combinations of code that
    loopy can generate.

    Objects of this type must be picklable.
    """

    # {{{ hashing/equality

    hash_fields: ClassVar[tuple[str, ...]] = ()
    comparison_fields: ClassVar[tuple[str, ...]] = ()

    def __hash__(self):
        # NOTE: _hash_value may vanish during pickling
        if getattr(self, "_hash_value", None) is None:
            from loopy.tools import LoopyKeyBuilder
            key_hash = LoopyKeyBuilder.new_hash()
            LoopyKeyBuilder()(self)
            object.__setattr__(self, "_hash_value", hash(key_hash.digest()))

        return self._hash_value  # pylint: disable=no-member

    def update_persistent_hash(self, key_hash, key_builder):
        key_hash.update(type(self).__name__.encode())
        for field_name in self.hash_fields:
            key_builder.rec(key_hash, getattr(self, field_name))

    def __eq__(self, other):
        if type(self) is not type(other):
            return False

        for field_name in self.comparison_fields:
            if getattr(self, field_name) != getattr(other, field_name):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    # }}}

    # {{{ preprocess

    def preprocess(self, kernel):
        return kernel

    def pre_codegen_entrypoint_check(self, kernel, callables_table):
        pass

    def pre_codegen_callable_check(self, kernel, callables_table):
        pass

    # }}}

    host_program_name_prefix = ""
    host_program_name_suffix = "_outer"
    device_program_name_prefix = ""
    device_program_name_suffix = ""

    def split_kernel_at_global_barriers(self) -> bool:
        """
        :returns: a :class:`bool` indicating whether the kernel should
            be split when a global barrier is encountered.
        """
        raise NotImplementedError()

    def get_host_ast_builder(self) -> ASTBuilderBase[Any]:
        """
        :returns: a class implementing :class:`ASTBuilderBase` for the host code
        """
        raise NotImplementedError()

    def get_device_ast_builder(self) -> ASTBuilderBase[Any]:
        """
        :returns: a class implementing :class:`ASTBuilderBase` for the host code
        """
        raise NotImplementedError()

    # {{{ types

    def get_dtype_registry(self):
        raise NotImplementedError()

    def is_vector_dtype(self, dtype):
        raise NotImplementedError()

    def vector_dtype(self, base, count):
        raise NotImplementedError()

    def alignment_requirement(self, type_decl):
        import struct
        return struct.calcsize(type_decl.struct_format())

    # }}}

    def get_kernel_executor_cache_key(self, *args, **kwargs):
        """
        :returns: an immutable type to be used as the cache key for
            kernel executor caching.
        """
        raise NotImplementedError()

    def get_kernel_executor(
            self, t_unit: TranslationUnit, *args, entrypoint: FunctionIdT,
            **kwargs) -> ExecutorBase:
        """
        :returns: an immutable type to be used as the cache key for
            kernel executor caching.
        """
        raise NotImplementedError()


class ASTBuilderBase(Generic[ASTType]):
    """An interface for generating (host or device) ASTs.
    """

    def __init__(self, target) -> None:
        self.target = target

    # {{{ library

    @property
    def known_callables(self):
        """
        Returns a mapping from function ids to corresponding
        :class:`loopy.kernel.function_interface.InKernelCallable` for the
        function ids known to *self.target*.
        """
        # FIXME: @inducer: Do we need to move this to TargetBase?
        return {}

    def symbol_manglers(self):
        return []

    def preamble_generators(self):
        return []

    # }}}

    # {{{ code generation guts

    @property
    def ast_module(self):
        raise NotImplementedError()

    def get_function_definition(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult,
            schedule_index: int, function_decl: ASTType, function_body: ASTType
            ) -> ASTType:
        raise NotImplementedError

    def get_function_declaration(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult, schedule_index: int
            ) -> tuple[Sequence[tuple[str, str]], ASTType | None]:
        """Returns preambles and the AST for the function declaration."""
        raise NotImplementedError

    def generate_top_of_body(
            self, codegen_state: CodeGenerationState) -> Sequence[ASTType]:
        return []

    def get_temporary_decls(self, codegen_state: CodeGenerationState,
            schedule_index: int) -> ASTType:
        raise NotImplementedError

    def get_kernel_call(self, codegen_state: CodeGenerationState,
            subkernel_name: str,
            gsize: tuple[Expression, ...],
            lsize: tuple[Expression, ...]) -> ASTType | None:
        raise NotImplementedError()

    @property
    def ast_block_class(self):
        raise NotImplementedError()

    @property
    def ast_block_scope_class(self):
        raise NotImplementedError()

    def get_expression_to_code_mapper(self, codegen_state: CodeGenerationState):
        raise NotImplementedError()

    def add_vector_access(self, access_expr, index):
        raise NotImplementedError()

    def emit_barrier(self, synchronization_kind, mem_kind, comment):
        """
        :arg synchronization_kind: ``"local"`` or ``"global"``
        :arg mem_kind: ``"local"`` or ``"global"``
        """
        raise NotImplementedError()

    def emit_assignment(self, codegen_state, insn):
        raise NotImplementedError()

    def emit_multiple_assignment(self, codegen_state, insn):
        raise NotImplementedError()

    def emit_sequential_loop(self, codegen_state, iname, iname_dtype,
            static_lbound, static_ubound, inner, hints):
        raise NotImplementedError()

    def emit_unroll_hint(self, value):
        raise NotImplementedError()

    @property
    def can_implement_conditionals(self):
        return False

    def emit_if(self, condition_str, ast):
        raise NotImplementedError()

    def emit_initializer(self, codegen_state, dtype, name, val_str, is_const):
        raise NotImplementedError()

    def emit_declaration_scope(self, codegen_state, inner):
        raise NotImplementedError()

    def emit_blank_line(self):
        raise NotImplementedError()

    def emit_comment(self, s):
        raise NotImplementedError()

    def emit_noop_with_comment(self, s):
        raise NotImplementedError()

    # }}}

    def process_ast(self, node):
        return node


# {{{ dummy host ast builder

class _DummyExpressionToCodeMapper:
    def rec(self, expr, prec, type_context=None, needed_dtype=None):
        return ""

    __call__ = rec


class _DummyASTBlock:
    def __init__(self, arg):
        self.contents = []

    def __str__(self):
        return ""


class DummyHostASTBuilder(ASTBuilderBase[None]):
    def get_function_definition(self, codegen_state, codegen_result,
            schedule_index, function_decl, function_body):
        return function_body

    def get_function_declaration(
            self, codegen_state, codegen_result,
            schedule_index,
            ) -> tuple[Sequence[tuple[str, str]], None]:
        return [], None

    def get_temporary_decls(self, codegen_state, schedule_index):
        return []

    def get_expression_to_code_mapper(self, codegen_state):
        return _DummyExpressionToCodeMapper()

    def get_kernel_call(self, codegen_state, name, gsize, lsize):
        return None

    @property
    def ast_block_class(self):
        return _DummyASTBlock

    @property
    def ast_block_scope_class(self):
        return _DummyASTBlock

# }}}


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2020 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

from pytools import memoize_method

from loopy.kernel import LoopKernel
from loopy.schedule import (
    Barrier,
    BeginBlockItem,
    CallKernel,
    EnterLoop,
    LeaveLoop,
    ReturnFromKernel,
    ScheduleItem,
    gather_schedule_block,
)


if TYPE_CHECKING:
    import loopy.kernel.data
    from loopy.kernel.instruction import InstructionBase


__doc__ = """
.. autoclass:: KernelProxyForCodegenOperationCacheManager

.. autoclass:: CodegenOperationCacheManager
"""


@dataclass(frozen=True)
class KernelProxyForCodegenOperationCacheManager:
    """
    Proxy to :class:`loopy.LoopKernel` to be used by
    :class:`CodegenOperationCacheManager`.
    """
    instructions: list[InstructionBase]
    linearization: list[ScheduleItem]
    inames: dict[str, loopy.kernel.data.Iname]

    @cached_property
    def id_to_insn(self):
        return {insn.id: insn for insn in self.instructions}

    def insn_inames(self, insn):
        return LoopKernel.insn_inames(self, insn)

    def iname_tags_of_type(self, iname, tag_type_or_types):
        from loopy.kernel.data import filter_iname_tags_by_type
        return filter_iname_tags_by_type(
                self.inames[iname].tags,
                tag_type_or_types)

    def __eq__(self, other):
        if not isinstance(other, (LoopKernel,
                KernelProxyForCodegenOperationCacheManager)):
            return False

        # TODO: could be more precise by only looking the inames' attributes
        # relevant to CodegenOperationCacheManager
        return (self.inames == other.inames
                and self.instructions == other.instructions
                and self.linearization == other.linearization)


class CodegenOperationCacheManager:
    """
    Caches operations arising during the codegen pipeline.

    .. attribute:: kernel_proxy

        An instance of :class:`KernelProxyForCodegenOperationCacheManager`.

    .. automethod:: with_kernel
    .. automethod:: get_concurrent_inames_in_a_callkernel
    """
    def __init__(self, kernel_proxy):
        assert isinstance(kernel_proxy, KernelProxyForCodegenOperationCacheManager)
        self.kernel_proxy = kernel_proxy

    @staticmethod
    def from_kernel(kernel):
        assert isinstance(kernel, LoopKernel)
        return CodegenOperationCacheManager(
                KernelProxyForCodegenOperationCacheManager(kernel.instructions,
                            kernel.linearization,
                            kernel.inames))

    def with_kernel(self, kernel):
        """
        Returns a new instance of :class:`CodegenOperationCacheManager`
        corresponding to *kernel* if the cached variables in *self* would
        be invalid for *kernel*, else returns *self*.
        """
        if self.kernel_proxy != kernel:
            # cached values are invalidated, must create a new one
            return CodegenOperationCacheManager.from_kernel(kernel)

        return self

    @cached_property
    def active_inames(self):
        """
        Returns an instance of :class:`list`, with the i-th entry being a
        :class:`frozenset` of active inames at the point just before i-th schedule
        item.
        """
        active_inames = []

        for i in range(len(self.kernel_proxy.linearization)):
            if i == 0:
                active_inames.append(frozenset())
                continue

            sched_item = self.kernel_proxy.linearization[i-1]
            if isinstance(sched_item, EnterLoop):
                active_inames.append(active_inames[i-1]
                        | frozenset([sched_item.iname]))
            elif isinstance(sched_item, LeaveLoop):
                assert sched_item.iname in active_inames[i-1]
                active_inames.append(active_inames[i-1]
                                     - frozenset([sched_item.iname]))
            else:
                active_inames.append(active_inames[i-1])

        return active_inames

    @cached_property
    def callkernel_index(self):
        """
        Returns an instance of :class:`list`, with the i-th entry being the index of
        :class:`loopy.schedule.CallKernel` containing the point just before i-th
        schedule item.
        """
        callkernel_index = []

        for i in range(len(self.kernel_proxy.linearization)):
            if i == 0:
                callkernel_index.append(None)
                continue

            sched_item = self.kernel_proxy.linearization[i-1]

            if isinstance(sched_item, CallKernel):
                callkernel_index.append(i-1)
            elif isinstance(sched_item, ReturnFromKernel):
                callkernel_index.append(None)
            else:
                callkernel_index.append(callkernel_index[i-1])

        return callkernel_index

    @cached_property
    def has_barrier_within(self):
        """
        Returns an instance of :class:`list`. The list's i-th entry is *True* if the
        i-th schedule item is a :class:`loopy.schedule.BeginBlockItem` containing a
        barrier or if the i-th schedule item is a :class:`loopy.schedule.Barrier`.
        """
        has_barrier_within = []

        for sched_idx, sched_item in enumerate(self.kernel_proxy.linearization):
            if isinstance(sched_item, BeginBlockItem):
                # TODO: calls to "gather_schedule_block" can be amortized
                _, endblock_index = gather_schedule_block(self.kernel_proxy
                                                          .linearization,
                                                          sched_idx)
                has_barrier_within.append(any(
                        isinstance(self.kernel_proxy.linearization[i], Barrier)
                        for i in range(sched_idx+1, endblock_index)))
            elif isinstance(sched_item, Barrier):
                has_barrier_within.append(True)
            else:
                has_barrier_within.append(False)

        return has_barrier_within

    @memoize_method
    def get_insn_ids_for_block_at(self, sched_index):
        """
        Cached variant of :func:`loopy.schedule.get_insn_ids_for_block_at`.
        """
        from loopy.schedule import get_insn_ids_for_block_at
        return get_insn_ids_for_block_at(self.kernel_proxy.linearization,
                                         sched_index)

    @memoize_method
    def get_concurrent_inames_in_a_callkernel(
            self, callkernel_index: int) -> frozenset[str]:
        """
        Returns a :class:`frozenset` of concurrent inames in a callkernel

        :arg callkernel_index: Index of the :class:`loopy.schedule.CallKernel`
            in the :attr:`CodegenOperationCacheManager.kernel_proxy`'s
            schedule, whose parallel inames are to be found.
        """

        from loopy.kernel.data import ConcurrentTag
        assert isinstance(self.kernel_proxy.linearization[callkernel_index],
                          CallKernel)
        insn_ids_in_subkernel = self.get_insn_ids_for_block_at(callkernel_index)

        inames_in_subkernel = {iname
                               for insn in insn_ids_in_subkernel
                               for iname in self.kernel_proxy.insn_inames(insn)}

        return frozenset([iname
                          for iname in inames_in_subkernel
                          if self.kernel_proxy.iname_tags_of_type(iname,
                                                                  ConcurrentTag)])


# vim: fdm=marker



import numpy as np

import pyopencl as cl
import pyopencl.array

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


# setup
# -----
ctx = cl.create_some_context()
queue = cl.CommandQueue(ctx)

n = 15 * 10**6
a = cl.array.arange(queue, n, dtype=np.float32)

# create
# ------
knl = lp.make_kernel(
        "{ [i]: 0<=i<n }",
        "out[i] = 2*a[i]")

# transform
# ---------
knl = lp.split_iname(knl, "i", 128, outer_tag="g.0", inner_tag="l.0")

# execute
# -------
# easy, slower:
evt, (out,) = knl(queue, a=a)
# efficient, with caching:
knl_ex = knl.executor(ctx)
evt, (out,) = knl_ex(queue, a=a)
# ENDEXAMPLE

knl = lp.add_and_infer_dtypes(knl, {"a": np.dtype(np.float32)})
print(lp.generate_code_v2(knl).device_code())



__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import numpy as np
import pytest

import pyopencl as cl
import pyopencl.clrandom
from pyopencl.tools import (  # noqa: F401
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)

import loopy as lp


pytest.importorskip("fparser")
logger = logging.getLogger(__name__)


def test_fp_prec_comparison():
    # FIXME: This test should succeed even when the number is exactly
    # representable in single precision.
    #
    # https://gitlab.tiker.net/inducer/loopy/issues/187

    fortran_src_dp = """
        subroutine assign_scalar(a)
          real*8 a(1)

          a(1) = 1.1d0
        end
        """

    prg_dp = lp.parse_fortran(fortran_src_dp)

    fortran_src_sp = """
        subroutine assign_scalar(a)
          real*8 a(1)

          a(1) = 1.1
        end
        """

    prg_sp = lp.parse_fortran(fortran_src_sp)

    assert prg_sp != prg_dp


def test_assign_double_precision_scalar(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    fortran_src = """
        subroutine assign_scalar(a)
          real*8 a(1)

          a(1) = 1.1d0
        end
        """

    t_unit = lp.parse_fortran(fortran_src)
    print(lp.generate_code_v2(t_unit).device_code())
    assert "1.1;" in lp.generate_code_v2(t_unit).device_code()

    a_dev = cl.array.empty(queue, 1, dtype=np.float64, order="F")
    t_unit(queue, a=a_dev)

    abs_err = abs(a_dev.get()[0] - 1.1)
    assert abs_err < 1e-15


def test_assign_double_precision_scalar_as_rational(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    fortran_src = """
        subroutine assign_scalar(a)
          real*8 a(1)

          a(1) = 11
          a(1) = a(1) / 10
        end
        """

    t_unit = lp.parse_fortran(fortran_src)

    a_dev = cl.array.empty(queue, 1, dtype=np.float64, order="F")
    t_unit(queue, a=a_dev)

    abs_err = abs(a_dev.get()[0] - 1.1)
    assert abs_err < 1e-15


def test_assign_single_precision_scalar(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    fortran_src = """
        subroutine assign_scalar(a)
          real*8 a(1)

          a(1) = 1.1
        end
        """

    t_unit = lp.parse_fortran(fortran_src)

    import re
    assert re.search(r"1.1000000[0-9]*f", lp.generate_code_v2(t_unit).device_code())

    a_dev = cl.array.empty(queue, 1, dtype=np.float64, order="F")
    t_unit(queue, a=a_dev)

    abs_err = abs(a_dev.get()[0] - 1.1)
    assert abs_err > 1e-15
    assert abs_err < 1e-6


def test_fill(ctx_factory):
    fortran_src = """
        subroutine fill(out, a, n)
          implicit none

          real*8 a, out(n)
          integer n, i

          do i = 1, n
            out(i) = a
          end do
        end

        !$loopy begin
        !
        ! fill = lp.parse_fortran(SOURCE)
        ! fill = lp.split_iname(fill, "i", split_amount,
        !     outer_tag="g.0", inner_tag="l.0")
        ! RESULT = fill
        !
        !$loopy end
        """

    knl = lp.parse_transformed_fortran(fortran_src,
            pre_transform_code="split_amount = 128")

    assert "i_inner" in knl["fill"].all_inames()

    ctx = ctx_factory()

    lp.auto_test_vs_ref(knl, ctx, knl, parameters={"n": 5, "a": 5})


def test_fill_const(ctx_factory):
    fortran_src = """
        subroutine fill(out, a, n)
          implicit none

          real*8 a, out(n)
          integer n, i

          do i = 1, n
            out(i) = 3.45
          end do
        end
        """

    knl = lp.parse_fortran(fortran_src)

    ctx = ctx_factory()

    lp.auto_test_vs_ref(knl, ctx, knl, parameters={"n": 5, "a": 5})


def test_asterisk_in_shape(ctx_factory):
    fortran_src = """
        subroutine fill(out, out2, inp, n)
          implicit none

          real*8 a, out(n), out2(n), inp(*)
          integer n, i

          do i = 1, n
            a = inp(n)
            out(i) = 5*a
            out2(i) = 6*a
          end do
        end
        """

    knl = lp.parse_fortran(fortran_src)

    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    knl(queue, inp=np.array([1, 2, 3.]), n=3)


def test_assignment_to_subst(ctx_factory):
    fortran_src = """
        subroutine fill(out, out2, inp, n)
          implicit none

          real*8 a, out(n), out2(n), inp(n)
          integer n, i

          do i = 1, n
            a = inp(i)
            out(i) = 5*a
            out2(i) = 6*a
          end do
        end
        """

    knl = lp.parse_fortran(fortran_src)

    ref_knl = knl

    knl = lp.assignment_to_subst(knl, "a", "i")

    ctx = ctx_factory()
    lp.auto_test_vs_ref(ref_knl, ctx, knl, parameters={"n": 5})


def test_assignment_to_subst_two_defs(ctx_factory):
    fortran_src = """
        subroutine fill(out, out2, inp, n)
          implicit none

          real*8 a, out(n), out2(n), inp(n)
          integer n, i

          do i = 1, n
            a = inp(i)
            out(i) = 5*a
            a = 3*inp(n)
            out2(i) = 6*a
          end do
        end
        """

    knl = lp.parse_fortran(fortran_src)

    ref_knl = knl

    knl = lp.assignment_to_subst(knl, "a")

    ctx = ctx_factory()
    lp.auto_test_vs_ref(ref_knl, ctx, knl, parameters={"n": 5})


def test_assignment_to_subst_indices(ctx_factory):
    fortran_src = """
        subroutine fill(out, out2, inp, n)
          implicit none

          real*8 a(n), out(n), out2(n), inp(n)
          integer n, i

          do i = 1, n
            a(i) = 6*inp(i)
          enddo

          do i = 1, n
            out(i) = 5*a(i)
          end do
        end
        """

    knl = lp.parse_fortran(fortran_src)

    knl = lp.fix_parameters(knl, n=5)

    ref_knl = knl

    assert "a" in knl["fill"].temporary_variables
    knl = lp.assignment_to_subst(knl, "a")
    assert "a" not in knl["fill"].temporary_variables

    ctx = ctx_factory()
    lp.auto_test_vs_ref(ref_knl, ctx, knl)


def test_if(ctx_factory):
    fortran_src = """
        subroutine fill(out, out2, inp, n)
          implicit none

          real*8 a, b, out(n), out2(n), inp(n)
          integer n, i, j

          do i = 1, n
            a = inp(i)
            if (a.ge.3) then
                b = 2*a
                do j = 1,3
                    b = 3 * b
                end do
                out(i) = 5*b
            else
                out(i) = 4*a
            endif
          end do
        end
        """

    knl = lp.parse_fortran(fortran_src)

    ref_knl = knl

    knl = lp.assignment_to_subst(knl, "a")

    ctx = ctx_factory()
    lp.auto_test_vs_ref(ref_knl, ctx, knl, parameters={"n": 5})


def test_tagged(ctx_factory):
    fortran_src = """
        subroutine rot_norm(out, alpha, out2, inp, inp2, n)
          implicit none
          real*8 a, b, r, out(n), out2(n), inp(n), inp2(n)
          real*8 alpha
          integer n, i

          do i = 1, n
            !$loopy begin tagged: input
            a = cos(alpha)*inp(i) + sin(alpha)*inp2(i)
            b = -sin(alpha)*inp(i) + cos(alpha)*inp2(i)
            !$loopy end tagged: input

            r = sqrt(a**2 + b**2)
            a = a/r
            b = b/r

            out(i) = a
            out2(i) = b
          end do
        end
        """

    knl = lp.parse_fortran(fortran_src)

    assert sum(1 for insn in lp.find_instructions(knl, "tag:input")) == 2


@pytest.mark.parametrize("buffer_inames", [
    "",
    "i_inner,j_inner",
    ])
def test_matmul(ctx_factory, buffer_inames):
    ctx = ctx_factory()

    if (buffer_inames and
            ctx.devices[0].platform.name == "Portable Computing Language"):
        pytest.skip("crashes on pocl")

    fortran_src = """
        subroutine dgemm(m,n,ell,a,b,c)
          implicit none
          real*8 a(m,ell),b(ell,n),c(m,n)
          integer m,n,k,i,j,ell

          do j = 1,n
            do i = 1,m
              do k = 1,ell
                c(i,j) = c(i,j) + b(k,j)*a(i,k)
              end do
            end do
          end do
        end subroutine
        """

    prog = lp.parse_fortran(fortran_src)

    assert len(prog["dgemm"].domains) == 1

    ref_prog = prog

    prog = lp.split_iname(prog, "i", 16,
            outer_tag="g.0", inner_tag="l.1")
    prog = lp.split_iname(prog, "j", 8,
            outer_tag="g.1", inner_tag="l.0")
    prog = lp.split_iname(prog, "k", 32)
    prog = lp.assume(prog, "n mod 32 = 0")
    prog = lp.assume(prog, "m mod 32 = 0")
    prog = lp.assume(prog, "ell mod 16 = 0")

    prog = lp.extract_subst(prog, "a_acc", "a[i1,i2]", parameters="i1, i2")
    prog = lp.extract_subst(prog, "b_acc", "b[i1,i2]", parameters="i1, i2")
    prog = lp.precompute(prog, "a_acc", "k_inner,i_inner",
            precompute_outer_inames="i_outer, j_outer, k_outer",
            default_tag="l.auto")
    prog = lp.precompute(prog, "b_acc", "j_inner,k_inner",
            precompute_outer_inames="i_outer, j_outer, k_outer",
            default_tag="l.auto")

    prog = lp.buffer_array(prog, "c", buffer_inames=buffer_inames,
            init_expression="0", store_expression="base+buffer")

    lp.auto_test_vs_ref(ref_prog, ctx, prog,
                        parameters={"n": 128, "m": 128, "ell": 128})


@pytest.mark.xfail
def test_batched_sparse():
    fortran_src = """
        subroutine sparse(rowstarts, colindices, values, m, n, nvecs, nvals, x, y)
          implicit none

          integer rowstarts(m+1), colindices(nvals)
          real*8 values(nvals)
          real*8 x(n, nvecs), y(n, nvecs), rowsum(nvecs)

          integer m, n, rowstart, rowend, length, nvals, nvecs
          integer i, j, k

          do i = 1, m
            rowstart = rowstarts(i)
            rowend = rowstarts(i+1)
            length = rowend - rowstart

            do k = 1, nvecs
              rowsum(k) = 0
            enddo
            do k = 1, nvecs
              do j = 1, length
                rowsum(k) = rowsum(k) + &
                  x(colindices(rowstart+j-1),k)*values(rowstart+j-1)
              end do
            end do
            do k = 1, nvecs
              y(i,k) = rowsum(k)
            end do
          end do
        end

        """

    knl = lp.parse_fortran(fortran_src)

    knl = lp.split_iname(knl, "i", 128)
    knl = lp.tag_inames(knl, {"i_outer": "g.0"})
    knl = lp.tag_inames(knl, {"i_inner": "l.0"})
    knl = lp.add_prefetch(knl, "values",
            default_tag="l.auto")
    knl = lp.add_prefetch(knl, "colindices",
            default_tag="l.auto")
    knl = lp.fix_parameters(knl, nvecs=4)


def test_fuse_kernels(ctx_factory):
    fortran_template = """
        subroutine {name}(nelements, ndofs, result, d, q)
          implicit none
          integer e, i, j, k
          integer nelements, ndofs
          real*8 result(nelements, ndofs, ndofs)
          real*8 q(nelements, ndofs, ndofs)
          real*8 d(ndofs, ndofs)
          real*8 prev

          do e = 1,nelements
            do i = 1,ndofs
              do j = 1,ndofs
                do k = 1,ndofs
                  {inner}
                end do
              end do
            end do
          end do
        end subroutine
        """

    xd_line = """
        prev = result(e,i,j)
        result(e,i,j) = prev + d(i,k)*q(e,i,k)
        """
    yd_line = """
        prev = result(e,i,j)
        result(e,i,j) = prev + d(i,k)*q(e,k,j)
        """

    xderiv = lp.parse_fortran(
            fortran_template.format(inner=xd_line, name="xderiv"))
    yderiv = lp.parse_fortran(
            fortran_template.format(inner=yd_line, name="yderiv"))
    xyderiv = lp.parse_fortran(
            fortran_template.format(
                inner=(xd_line + "\n" + yd_line), name="xyderiv"))

    knl = lp.fuse_kernels((xderiv["xderiv"], yderiv["yderiv"]),
            data_flow=[("result", 0, 1)])
    knl = knl.with_kernel(lp.prioritize_loops(knl["xderiv_and_yderiv"], "e,i,j,k"))

    assert len(knl["xderiv_and_yderiv"].temporary_variables) == 2

    ctx = ctx_factory()
    lp.auto_test_vs_ref(xyderiv, ctx, knl, parameters={"nelements": 20, "ndofs": 4})


def test_parse_and_fuse_two_kernels():
    fortran_src = """
        subroutine fill(out, a, n)
          implicit none

          real*8 a, out(n)
          integer n, i

          do i = 1, n
            out(i) = a
          end do
        end

        subroutine twice(out, n)
          implicit none

          real*8 out(n)
          integer n, i

          do i = 1, n
            out(i) = 2*out(i)
          end do
        end

        !$loopy begin
        !
        ! t_unit = lp.parse_fortran(SOURCE)
        ! fill = t_unit["fill"]
        ! twice = t_unit["twice"]
        ! knl = lp.fuse_kernels((fill, twice))
        ! print(knl)
        ! RESULT = knl
        !
        !$loopy end
        """

    lp.parse_transformed_fortran(fortran_src)


def test_precompute_some_exist(ctx_factory):
    fortran_src = """
        subroutine dgemm(m,n,ell,a,b,c)
          implicit none
          real*8 a(m,ell),b(ell,n),c(m,n)
          integer m,n,k,i,j,ell

          do j = 1,n
            do i = 1,m
              do k = 1,ell
                c(i,j) = c(i,j) + b(k,j)*a(i,k)
              end do
            end do
          end do
        end subroutine
        """

    knl = lp.parse_fortran(fortran_src)

    assert len(knl["dgemm"].domains) == 1

    knl = lp.split_iname(knl, "i", 8,
            outer_tag="g.0", inner_tag="l.1")
    knl = lp.split_iname(knl, "j", 8,
            outer_tag="g.1", inner_tag="l.0")
    knl = lp.split_iname(knl, "k", 8)
    knl = lp.assume(knl, "n mod 8 = 0")
    knl = lp.assume(knl, "m mod 8 = 0")
    knl = lp.assume(knl, "ell mod 8 = 0")

    knl = lp.extract_subst(knl, "a_acc", "a[i1,i2]", parameters="i1, i2")
    knl = lp.extract_subst(knl, "b_acc", "b[i1,i2]", parameters="i1, i2")
    knl = lp.precompute(knl, "a_acc", "k_inner,i_inner",
            precompute_inames="ktemp,itemp",
            precompute_outer_inames="i_outer, j_outer, k_outer",
            default_tag="l.auto")
    knl = lp.precompute(knl, "b_acc", "j_inner,k_inner",
            precompute_inames="itemp,k2temp",
            precompute_outer_inames="i_outer, j_outer, k_outer",
            default_tag="l.auto")

    ref_knl = knl

    ctx = ctx_factory()
    lp.auto_test_vs_ref(ref_knl, ctx, knl, parameters={"n": 128, "m": 128, "ell": 128})


def test_fortran_subroutines():
    fortran_src = """
        subroutine twice(n, a)
          implicit none
          real*8  a(n)
          integer i,n

          do i=1,n
            a(i) = a(i) * 2
          end do
        end subroutine

        subroutine twice_cross(n, a, i)
          implicit none
          integer i, n
          real*8  a(n,n)

          call twice(n, a(1:n, i))
          call twice(n, a(i, 1:n))
        end subroutine
        """
    t_unit = lp.parse_fortran(fortran_src).with_entrypoints("twice_cross")
    print(lp.generate_code_v2(t_unit).device_code())


def test_domain_fusion_imperfectly_nested():
    fortran_src = """
        subroutine imperfect(n, m, a, b)
            implicit none
            integer i, j, n, m
            real a(n), b(n,n)

            do i=1, n
                a(i) = i
                do j=1, m
                    b(i,j) = i*j
                end do
            end do
        end subroutine
        """

    t_unit = lp.parse_fortran(fortran_src)
    # If n > 0 and m == 0, a single domain would be empty,
    # leading (incorrectly) to no assignments to 'a'.
    assert len(t_unit["imperfect"].domains) > 1


def test_division_in_shapes(ctx_factory):
    fortran_src = """
        subroutine halve(m, a)
            implicit none
            integer m, i, j
            real*8 a(m/2,m/2)
            do i = 1,m/2
                do j = 1,m/2
                    a(i, j) = 2*a(i, j)
                end do
            end do
        end subroutine
        """
    t_unit = lp.parse_fortran(fortran_src)
    ref_t_unit = t_unit

    print(t_unit)

    ctx = ctx_factory()
    lp.auto_test_vs_ref(ref_t_unit, ctx, t_unit, parameters={"m": 128})


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import islpy as isl

from loopy.kernel.data import ArrayArg, ValueArg
from loopy.symbolic import RuleAwareIdentityMapper, SubstitutionRuleMappingContext
from loopy.translation_unit import for_each_kernel


__doc__ = """
.. currentmodule:: loopy

.. autofunction:: to_batched
"""


# {{{ to_batched

def temp_needs_batching_if_not_sequential(tv, batch_varying_args):
    from loopy.kernel.data import AddressSpace
    if tv.name in batch_varying_args:
        return True
    if tv.initializer is not None and tv.read_only:
        # do not batch read_only temps  if not in
        # `batch_varying_args`
        return False
    if tv.address_space == AddressSpace.PRIVATE:
        # do not batch private temps if not in `batch_varying args`
        return False
    return True


class _BatchVariableChanger(RuleAwareIdentityMapper):
    def __init__(self, rule_mapping_context, kernel, batch_varying_args,
            batch_iname_expr, sequential):
        super().__init__(rule_mapping_context)

        self.kernel = kernel
        self.batch_varying_args = batch_varying_args
        self.batch_iname_expr = batch_iname_expr
        self.sequential = sequential

    def needs_batch_subscript(self, name):
        tv = self.kernel.temporary_variables.get(name)

        if name in self.batch_varying_args:
            return True
        if not self.sequential:
            if tv is None:
                return False
            if not temp_needs_batching_if_not_sequential(tv,
                    self.batch_varying_args):
                return False

        return True

    def map_subscript(self, expr, expn_state):
        if not self.needs_batch_subscript(expr.aggregate.name):
            return super().map_subscript(expr, expn_state)

        idx = self.rec(expr.index, expn_state)
        if not isinstance(idx, tuple):
            idx = (idx,)

        return type(expr)(expr.aggregate, (self.batch_iname_expr, *idx))

    def map_variable(self, expr, expn_state):
        if not self.needs_batch_subscript(expr.name):
            return super().map_variable(expr, expn_state)

        return expr[self.batch_iname_expr]


def _add_unique_dim_name(name, dim_names):
    if dim_names is None:
        return dim_names

    from pytools import UniqueNameGenerator
    ng = UniqueNameGenerator(set(dim_names))
    return (ng(name), *tuple(dim_names))


@for_each_kernel
def to_batched(kernel, nbatches, batch_varying_args, batch_iname_prefix="ibatch",
        sequential=False):
    """Takes in a kernel that carries out an operation and returns a kernel
    that carries out a batch of these operations.

    .. note::
       For temporaries in a kernel that are private or read only
       globals and if `sequential=True`, loopy does not does not batch these
       variables unless explicitly mentioned in `batch_varying_args`.

    :arg nbatches: the number of batches. May be a constant non-negative
        integer or a string, which will be added as an integer argument.
    :arg batch_varying_args: a list of argument names that vary per-batch.
        Each such variable will have a batch index added.
    :arg sequential: A :class:`bool`. If *True*, do not duplicate
        temporary variables for each batch. This automatically tags the batch
        iname for sequential execution.
    """

    from pymbolic import var

    vng = kernel.get_var_name_generator()
    batch_iname = vng(batch_iname_prefix)
    batch_iname_expr = var(batch_iname)

    new_args = []

    batch_dom_str = "{{[{iname}]: 0 <= {iname} < {nbatches}}}".format(
            iname=batch_iname,
            nbatches=nbatches,
            )

    if not isinstance(nbatches, int):
        batch_dom_str = "[%s] -> " % nbatches + batch_dom_str
        new_args.append(ValueArg(nbatches, dtype=kernel.index_dtype))

        nbatches_expr = var(nbatches)
    else:
        nbatches_expr = nbatches

    batch_domain = isl.BasicSet(batch_dom_str)
    new_domains = [batch_domain, *kernel.domains]

    for arg in kernel.args:
        if arg.name in batch_varying_args:
            if isinstance(arg, ValueArg):
                arg = ArrayArg(arg.name, arg.dtype, shape=(nbatches_expr,),
                        dim_tags="c")
            else:
                arg = arg.copy(
                        shape=(nbatches_expr, *arg.shape),
                        dim_tags=("c",) * (len(arg.shape) + 1),
                        dim_names=_add_unique_dim_name("ibatch", arg.dim_names))

        new_args.append(arg)

    kernel = kernel.copy(
            domains=new_domains,
            args=new_args)

    if not sequential:
        new_temps = {}

        for temp in kernel.temporary_variables.values():
            if temp_needs_batching_if_not_sequential(temp, batch_varying_args):
                new_temps[temp.name] = temp.copy(
                        shape=(nbatches_expr, *temp.shape),
                        dim_tags=("c",) * (len(temp.shape) + 1),
                        dim_names=_add_unique_dim_name("ibatch", temp.dim_names))
            else:
                new_temps[temp.name] = temp

        kernel = kernel.copy(temporary_variables=new_temps)
    else:
        import loopy as lp
        from loopy.kernel.data import ForceSequentialTag
        kernel = lp.tag_inames(kernel, [(batch_iname, ForceSequentialTag())])

    rule_mapping_context = SubstitutionRuleMappingContext(
            kernel.substitutions, vng)
    bvc = _BatchVariableChanger(rule_mapping_context,
            kernel, batch_varying_args, batch_iname_expr,
            sequential=sequential)
    kernel = rule_mapping_context.finish_kernel(
            bvc.map_kernel(kernel))

    batch_iname_set = frozenset([batch_iname])
    kernel = kernel.copy(
            instructions=[
                insn.copy(within_inames=insn.within_inames | batch_iname_set)
                for insn in kernel.instructions])

    return kernel

# }}}

# vim: foldmethod=marker



from __future__ import annotations

import sys

import numpy as np

import loopy as lp


def to_python_literal(value):
    try:
        int(value)
    except ValueError:
        pass
    else:
        # It's an integer
        return value

    try:
        float(value)
    except ValueError:
        pass
    else:
        # It's a float
        return repr(float(value))

    if value.endswith("f"):
        try:
            float(value[:-1])
        except ValueError:
            pass
        else:
            # It's a float
            return repr(float(value[:-1]))

    return repr(value)


def defines_to_python_code(defines_str):
    import re
    define_re = re.compile(r"^\#define\s+([a-zA-Z0-9_]+)\s+(.*)$")
    result = []
    for line in defines_str.split("\n"):
        if not line.strip():
            continue

        match = define_re.match(line)
        if match is None:
            raise RuntimeError("#define not understood: '%s'" % line)

        result.append(
                "{} = {}".format(match.group(1), to_python_literal(match.group(2))))

    return "\n".join(result)


def main():
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Stand-alone loopy frontend")

    parser.add_argument("infile", metavar="INPUT_FILE")
    parser.add_argument("outfile", default="-", metavar="OUTPUT_FILE",
            help="Defaults to stdout ('-').", nargs="?")
    parser.add_argument("--lang", metavar="LANGUAGE", help="loopy|fortran")
    parser.add_argument("--target", choices=(
        "opencl", "ispc", "ispc-occa", "c", "c-fortran", "cuda"),
        default="opencl")
    parser.add_argument("--transform")
    parser.add_argument("--edit-code", action="store_true")
    parser.add_argument("--occa-defines")
    parser.add_argument("--print-ir", action="store_true")
    args = parser.parse_args()

    if args.target == "opencl":
        from loopy.target.opencl import OpenCLTarget
        target = OpenCLTarget
    elif args.target == "ispc":
        from loopy.target.ispc import ISPCTarget
        target = ISPCTarget
    elif args.target == "ispc-occa":
        from loopy.target.ispc import ISPCTarget
        target = lambda: ISPCTarget()  # noqa: E731
    elif args.target == "c":
        from loopy.target.c import CTarget
        target = CTarget
    elif args.target == "c-fortran":
        from loopy.target.c import CTarget
        target = lambda: CTarget(fortran_abi=True)  # noqa: E731
    elif args.target == "cuda":
        from loopy.target.cuda import CudaTarget
        target = CudaTarget
    else:
        raise ValueError(f"unknown target: {args.target}")

    lp.set_default_target(target)

    lang = None
    if args.infile == "-":
        infile_content = sys.stdin.read()
    else:
        from os.path import splitext
        _, ext = splitext(args.infile)

        lang = {
                ".py": "loopy",
                ".loopy": "loopy",
                ".floopy": "fortran",
                ".f90": "fortran",
                ".F90": "fortran",
                ".fpp": "fortran",
                ".f": "fortran",
                ".f77": "fortran",
                ".F77": "fortran",
                }.get(ext)
        with open(args.infile) as infile_fd:
            infile_content = infile_fd.read()

    if args.lang is not None:
        lang = args.lang

    if lang is None:
        raise RuntimeError("unable to deduce input language "
                "(wrong input file extension? --lang flag?)")

    if lang == "loopy":
        # {{{ path wrangling

        from os import getcwd
        from os.path import abspath, dirname

        infile_dirname = dirname(args.infile)
        if infile_dirname:
            infile_dirname = abspath(infile_dirname)
        else:
            infile_dirname = getcwd()

        sys.path.append(infile_dirname)

        # }}}

        data_dic = {}
        data_dic["lp"] = lp
        data_dic["np"] = np

        if args.occa_defines:
            with open(args.occa_defines) as defines_fd:
                occa_define_code = defines_to_python_code(defines_fd.read())
            exec(compile(occa_define_code, args.occa_defines, "exec"), data_dic)

        with open(args.infile) as infile_fd:
            exec(compile(infile_content, args.infile, "exec"), data_dic)

        if args.transform:
            with open(args.transform) as xform_fd:
                exec(compile(xform_fd.read(),
                    args.transform, "exec"), data_dic)

        try:
            kernel = data_dic["lp_knl"]
        except KeyError as err:
            raise RuntimeError("loopy-lang requires 'lp_knl' "
                    "to be defined on exit") from err

        t_unit = [kernel]

    elif lang in ["fortran", "floopy", "fpp"]:
        pre_transform_code = None
        if args.transform:
            with open(args.transform) as xform_fd:
                pre_transform_code = xform_fd.read()

        if args.occa_defines:
            if pre_transform_code is None:
                pre_transform_code = ""

            with open(args.occa_defines) as defines_fd:
                pre_transform_code = (
                        defines_to_python_code(defines_fd.read())
                        + pre_transform_code)

        t_unit = lp.parse_transformed_fortran(
                infile_content, pre_transform_code=pre_transform_code,
                filename=args.infile)

    else:
        raise RuntimeError("unknown language: '%s'"
                % args.lang)

    if not isinstance(t_unit, lp.TranslationUnit):
        # FIXME
        assert isinstance(t_unit, list)  # of kernels
        raise NotImplementedError("convert list of kernels to TranslationUnit")

    if args.print_ir:
        print(t_unit, file=sys.stderr)

    t_unit = lp.preprocess_kernel(t_unit)
    cgr = lp.generate_code_v2(t_unit)

    if args.outfile is not None:
        outfile = args.outfile
    else:
        outfile = "-"

    code = cgr.device_code()

    if outfile == "-":
        sys.stdout.write(code)
    else:
        with open(outfile, "w") as outfile_fd:
            outfile_fd.write(code)
